# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
from sqlalchemy import Table, Column, Integer, String
from sqlalchemy.orm.interfaces import AttributeExtension, MapperExtension, InstrumentationManager
from sqlalchemy.orm import object_session, EXT_STOP, EXT_CONTINUE, sessionmaker, session
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm.exc import UnmappedInstanceError
from sqlalchemy.dialects.postgresql import UUID as pg_UUID
import uuid
import inspect

class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    "Backend-agnostic GUID Type" from http://www.sqlalchemy.org/docs/core/types.html
    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(pg_UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value.hex

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)

def sessionExtension(**kwargs):
    """
    Class decorator function to indicate classes in modules which are :class:`ModelSession`  
    extensions.  See :func:`ModelSessionExtension` for details about parameters.
    """
    def apply(cls):
        cls.modelSessionExtension = ModelSessionExtension(**kwargs)
        return cls
    return apply

class ModelSessionExtension(object):
    def __init__(self, name=None, names=None, clsName=None, clsNames=None):
        if name is None:
            self.names = names
        elif names is not None:
            self.names = names
            self.names.append(name)
        else:
            self.names = [name]

        if clsName is None:
            self.clsNames = clsNames
        elif clsNames is not None:
            self.clsNames = clsNames
            self.clsNames.append(clsName)
        else:
            self.clsNames = [clsName]

    def matches(self, target):
        if self.names is not None and hasattr(target, "objectName") and target.objectName() in self.names:
            return True

        if self.clsNames is not None and target.__class__.__name__ in self.clsNames:
            return True

class ModelSession(session.Session):
    """
    The ModelSession extends the sqlalchemy Session base class with tools for 
    convenience object addition and objects which are truly not a part of the 
    database persistence model (although may become so).
    
    The return of :func:`PBSessionMaker` is a class derived from ModelSession.
    """

    def modelListeners(self, event):
        for l in targetInstrumentation(self).event_list(event):
            yield l
        for l in targetInstrumentation(self.__class__).event_list(event):
            yield l

    def modelMessage(self, instance, flags, msg):
        r = None
        for listener in self.modelListeners("message"):
            r = listener(self, instance, flags, msg)
            if r is not None:
                break
        return r

    def npadd(self, obj):
        """
        npadd is an add for non-persistent objects.  This could be objects that 
        are candidates to join the persistent store, but have not due to been 
        confirmed by the user.  Alternatively, objects which are merely UI 
        artifacts may be added here so that their UserAttr attributes know 
        which session instance should be used for database interaction and 
        local caching.
        """
        targetInstrumentation(obj).session = self

    @classmethod
    def _prepExtensionClasses(cls):
        if not hasattr(cls, "extensionClasses"):
            cls.extensionClasses = []

    @classmethod
    def addExtensionModule(cls, module):
        """
        Add a module extension with classes.  It is expected that dialogs and
        command class emit events using :func:`event` which can be hooked by
        classes in the passed module.
        """
        cls._prepExtensionClasses()
        for n in dir(module):
            if hasattr(getattr(module, n), "modelSessionExtension"):
                cls.extensionClasses.append(getattr(module, n))

    @classmethod
    def event(cls, evt, target, *params, **kwargs):
        """
        Call an event on the extension classes associated with this ModelSession
        instance.  Extensions must be added to the ModelSession instance with
        :func:`addExtensionModule`.
        """
        cls._prepExtensionClasses()
        for e in cls.extensionClasses:
            if e.modelSessionExtension.matches(target) and hasattr(e, evt):
                getattr(e, evt)(target, *params, **kwargs)

def PBSessionMaker(**kwargs):
    """
    This is qtalchemy alternative for sqlalchemy.orm.sessionmaker.  We 
    default a few parameters specially and pass it on.  The returned 
    class is derived from :class:`ModelSession`.
    
    The parameters are the same as sqlalchemy.orm.sessionmaker.
    """
    # we're a gui framework we want to write on dialog ok
    if not kwargs.has_key("autoflush"):
        kwargs["autoflush"] = False
    # we want failed commits (due to pre-save validation) in a bound dialog to live on for the second try
    if not kwargs.has_key("expire_on_commit"):
        kwargs["expire_on_commit"] = False
    # When you get a server side commit failure this flag keeps sqlalchemy from expiring my session items.
    if not kwargs.has_key("_enable_transaction_accounting"):
        kwargs["_enable_transaction_accounting"] = False
    # pass it on
    return sessionmaker(class_=ModelSession,**kwargs)

class InstallAttributeListeners(InstrumentationManager):
    def post_configure_attribute(self, class_, key, inst):
        """Add an event listener to an InstrumentedAttribute."""
        inst.impl.extensions.insert(0, ModelChangeExtension(key))

class ModelChangeExtension(AttributeExtension):
    """
    This extension will activate a call-back on the object.  This 
    call-back is expected to iterate through a list of listening 
    models and notify them of the changes.
    """
    def __init__(self, key):
        self.key = key
    
    def append(self, state, value, initiator):
        self._report(state, value, None, "appended")
        return value

    def remove(self, state, value, initiator):
        self._report(state, value, None, "removed")
    
    def _report(self, state, value, oldvalue, verb):
        state.obj().receive_change_event(verb, self.key, value, oldvalue)


class Message:
    Ok      = 0x000001
    Cancel  = 0x000002
    Yes     = 0x000004
    No      = 0x000008
    Information = 0x001000
    Warning     = 0x002000
    Abort       = 0x004000

    # no return implies None and, thus, ContinueProcessing
    ContinueProcessing = None

class EventExtension(MapperExtension):
    """
    Enables hooks on mapped SA objects with names like:

    __before_insert__
    __after_insert__
    etc.

    See SA documentation and the _add_handler code below for details.
    """
    def handle_event(self, event, mapper, connection, instance):
        name = '__%s__' % event
        handler = getattr(instance, name, None)
        if handler:
            handler()

def _add_handler(event):
    def handle_event(self, *args, **kwargs):
        self.handle_event(event, *args, **kwargs)
        return EXT_CONTINUE
    setattr(EventExtension, event, handle_event)
    
for _name in ['before_insert',
            'after_insert',
            'before_update',
            'after_update',
            'before_delete',
            'after_delete']:
    _add_handler(_name)

event_extension = EventExtension()

class ModelInstrumentation(object):
    def __init__(self):
        self.events = {}
        self.setting = {}
        
    def event_list(self, event, attribute=None):
        if attribute is None:
            if self.events.has_key(event):
                return self.events[event]
            else:
                return []
        else:
            x = "{0}.".format(event)
            y = "{0}.{1}".format(event, attribute)
            return self.event_list(x) + self.event_list(y)

    def event_add(self, event, attribute, listener):
        if event not in ["message", "check", "set", "get"]:
            raise ValueError("Event '{0}' is not a valid event.".format(event))

        if attribute is not None:
            event = "{0}.{1}".format(event, attribute)
        if self.events.has_key(event):
            self.events[event].append(listener)
        else:
            self.events[event] = [listener]

    def add(self, event, attribute=None):
        def decorate(f):
            self.event_add(event, attribute, f)
            return f
        return decorate

def targetInstrumentation(target):
    if inspect.isclass(target):
        try:
            return target.classEvents
        except AttributeError as e:
            target.classEvents = ModelInstrumentation()
            return target.classEvents
    else:
        try:
            return target._instrumentation
        except AttributeError as e:
            target._instrumentation = ModelInstrumentation()
            return target._instrumentation

def instanceEvent(target, event, attribute=None):
    def decorate(f):
        x=targetInstrumentation(target)
        x.event_add(event, attribute, f)
        return f
    return decorate

class ModelObject(object):
    """
    A class derived from ModelObject provides an event model to notify 
    :class:`qtalchemy.PBTableModel` when attributes are modified.

    The method names in this class are prefixed with 'model' to keep them 
    separate from the application programmers attributes.
    """
    # Add the mapped object events
    __mapper_args__ = {'extension': EventExtension()}
    # add the attribute events for model call backs
    # TODO:  InstallAttributeListeners is broken in sqlalchemy 0.7.0 
    # We currently don't have code relying on these events.  The set was just replaced 
    # and the collection append/remove have not been utilized.
    # TODO 1:  append/remove notifications should notify the model
    # TODO 2:  clean up these event hooks
    #__sa_instrumentation_manager__ = InstallAttributeListeners

    # this is merely a name-space trick to make things more memorable
    Events = ModelInstrumentation

    def session(self):
        """
        Obtain a pointer to session in which this object lives.

        TODO:  rename session to modelSession
        """
        sessionCandidate = None
        try:
            sessionCandidate = object_session(self)
        except UnmappedInstanceError as e:
            pass
        if sessionCandidate is None and hasattr(targetInstrumentation(self), "session"):
            sessionCandidate = targetInstrumentation(self).session
        return sessionCandidate

    def modelListeners(self, event, attribute=None):
        """
        Enumerate the listeners on a particular event.
        """
        for l in targetInstrumentation(self).event_list(event, attribute):
            yield l
        #TODO:  Consider navigating the entire MRO
        #           for cls in self.__class__.mro()
        for l in targetInstrumentation(self.__class__).event_list(event, attribute):
            yield l

    def modelMessage(self, flags, msg):
        """
        Call out to the 'message' event handlers of this object and the session.
        
        When connected to a UI, it is expected that the session message event 
        is connected to the :class:`qtalchemy.dialogs.SessionMessageBox` object 
        which displays the message in a Qt message box.
        
        :param flags: bit mask of flags from `Message`
        """
        r = None
        s = self.session()
        for listener in self.modelListeners("message"):
            r = listener(self, flags, msg)
            if r is not None:
                break
        if r is None and s is not None:
            r = s.modelMessage(self, flags, msg)
        return r

    def __setattr__(self, attr, v):
        if attr.startswith("_instrumentation") or attr.startswith("_UserAttr"):
            super(ModelObject,self).__setattr__(attr,v)
        else:
            try:
                oldvalue = self.__dict__[attr]
            except KeyError as e:
                oldvalue = None
            inst = targetInstrumentation(self)
            if not inst.setting.has_key(attr):
                inst.setting[attr] = 0
            assert 0 == inst.setting[attr], "Recursive set of {0}".format(attr)
            inst.setting[attr] += 1
            try:
                for listener in self.modelListeners("check", attr):
                    listener(self, attr, v)
                super(ModelObject,self).__setattr__(attr,v)
                for listener in self.modelListeners("set", attr):
                    listener(self, attr, oldvalue)
                self.receive_change_event("set",attr,v,None)
            finally:
                inst.setting[attr] -= 1

    def receive_change_event(self, verb, key, value, oldvalue):
        if not hasattr(self,"_ModelObject__listening_models"):
            return # no models to notify
        
        for m in self.__listening_models:
            # tell the model about this event
            m.row_object_change(self,verb,key,value,oldvalue)

    def is_setting(self,attr):
        """
        Check current assignment recursion flags.  Return True if the specified 
        attribute is currently being assigned.
        """
        inst = targetInstrumentation(self)
        return inst.setting.has_key(attr) and inst.setting[attr] > 0

class ValidationError(Exception):
    """
    This validation class is meant to be used from the validation events connected via ModelObject.
    """
    pass

def user_message(exception):
    """
    This helper function returns a polite user message derived from a sqlalchemy database exception.
    
    It probably does a fairly horrible job of it.
    """
    try:
        if isinstance(exception,OperationalError):
            return exception.orig.args[1]
        if isinstance(exception,ProgrammingError):
            return "%s\n\nOffending SQL Statement:  %s" % (exception.orig.args[1], exception.statement)
    except:
        pass
    return str(exception)



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
This module contains tools to enable cross platform support.
"""

from stderr import StdErrEater, guiexcepthook

from util import is_windows, xdg_open



# -*- coding: utf-8 -*-
"""
We create models, attach to data and see if the events all propogate.
"""

from qtalchemy import UserAttr, ClassTableModel, ModelObject
from PyQt4 import QtCore, QtGui

class ItemTest(ModelObject):
    classEvents = ModelObject.Events()

    def __init__(self,domain=""):
        self.domain = domain
    
    name = UserAttr(str,"Name")
    domain = UserAttr(str,"Domain")
    email = UserAttr(str,"E-Mail")

    @classEvents.add("set", "name")
    @classEvents.add("set", "domain")
    def on_change_email_parts(self, attr, oldvalue):
        self.email = "%s@%s" % (self.name, self.domain)

def create_model_test():
    domains = ["a.com", "b.com", "c.com"]
    personas = [ItemTest(domain=b) for b in domains]
    
    # objects referenced in model_data_change
    these_were_touched = []
    
    def model_data_change(index1,index2):
        these_were_touched.append(index1.internalPointer())
    
    model = ClassTableModel(ItemTest,"name,domain,email".split(','),readonly=False)
    model.dataChanged.connect(model_data_change)
    model.reset_content_from_list(personas)
    
    model.setData(model.index(0,0,None),"jeff",QtCore.Qt.EditRole)
    
    assert personas[0].name == "jeff", "persona 0 should be jeff"
    assert len(these_were_touched) >= 1, "should have gotten a notification for jeff's modification"
    these_were_touched = []
    
    for p in personas:
        p.name='joe'
        
    assert len(these_were_touched) >= 3, "should have gotten a bunch of notifications for joe changes"



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

"""
An input yoke provides knowledge of the Qt property names and signals 
associated with widget input.  This ties the input widgets to the model 
attributes.

The yokes work with a mapper object which holds the object whose attributes 
are being edited.
"""

from PyQtModels import *

class InputYoke(object):
    """
    InputYoke is an abstract base class.
    """
    
    class Label(object):
        External = 0
        Internal = 1
    
    def __init__(self,mapper):
        self.mapper = mapper

    def Factory(self):
        """
        In this event you should:
        
        #. Create the widget
        #. Set size hints and other widget attributes
        #. Connect to signals
        #. Return the newly created widet
        """
        return None

    def AdoptWidget(self, widget):
        self.widget = widget
        self._baseAdoptWidget(self.widget)

    def _baseAdoptWidget(self, widget):
        widget.setWhatsThis(ClassAttributeWhatsThis(getattr(self.mapper.cls, self.attr)))
        widget.setProperty("BoundClassField", "{0}.{1}".format(self.mapper.cls.__name__, self.attr))

    def focus(self):
        self.widget.setFocus(QtCore.Qt.OtherFocusReason)

    def mainWidget(self):
        return self.widget

    def Bind(self):
        """
        Data initialization.
        """
        pass

    def Update(self,key):
        self.Bind()

    def Save(self):
        """
        Data initialization.
        """
        pass

    def LabelStyle(self):
        return InputYoke.Label.External

class LineYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QtGui.QLineEdit()
        self.widget.editingFinished.connect(self.Save)
        self._baseAdoptWidget(self.widget)
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        self._baseAdoptWidget(self.widget)
        if isinstance(self.widget, QtGui.QLineEdit):
            self.widget.editingFinished.connect(self.Save)
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)

    def Bind(self):
        if isinstance(self.widget, QtGui.QLineEdit):
            self.widget.setText(self.mapper.getObjectAttr(self.attr))
        elif isinstance(self.widget, QtGui.QComboBox):
            if self.widget.isEditable():
                self.widget.setEditText(self.mapper.getObjectAttr(self.attr))
            else:
                self.widget.setCurrentIndex(self.widget.findText(self.mapper.getObjectAttr(self.attr), QtCore.Qt.MatchFixedString))
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        if isinstance(self.widget, QtGui.QLineEdit):
            self.mapper.setObjectAttr(self.attr,self.widget.text())
        elif isinstance(self.widget, QtGui.QComboBox):
            self.mapper.setObjectAttr(self.attr,self.widget.currentText())
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

class SelectionYoke(LineYoke):
    """
    This yoke drives combo boxes.  Note that this yoke is not directly 
    usable as a yoke factory due to the options parameter.  It is intended 
    to be wrapped in a factory function and exposed to the global yoke 
    factory with :func:`addGlobalYoke`.
    
    >>> def ColorComboYoke(mapper, attr):
    ...     return SelectionYoke(mapper, attr, ["Red", "Orange", "Yellow", "Green", "Blue", "Indigo", "Violet"])
    ...
    >>> addGlobalYoke('color', ColorComboYoke)
    
    :param options:  a list of strings to include in the combo box
    """
    def __init__(self, mapper, attr, options):
        LineYoke.__init__(self, mapper, attr)
        self.options = options

    def Factory(self):
        self.widget = QtGui.QComboBox()
        self._baseAdoptWidget(self.widget)
        self._loadOptions()
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        self._baseAdoptWidget(self.widget)
        self._loadOptions()
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)

    def _loadOptions(self):
        for o in self.options:
            self.widget.addItem(o)

class PasswordYoke(LineYoke):
    def Factory(self):
        LineYoke.Factory(self)
        self.widget.setEchoMode(QtGui.QLineEdit.Password)
        return self.widget

    def AdoptWidget(self, widget):
        LineYoke.AdoptWidget(self, widget)
        self.widget.setEchoMode(QtGui.QLineEdit.Password)

class BlankIntValidator(QtGui.QIntValidator):
    def validate(self,input,pos):
        if input == "":
            return QtGui.QValidator.Acceptable, input, pos
        else:
            return QtGui.QIntValidator.validate(self,input,pos)

class IntegerYoke(LineYoke):
    def Factory(self):
        self.widget = QtGui.QLineEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.setValidator(BlankIntValidator())
        self.widget.editingFinished.connect(self.Save)
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def Bind(self):
        x = self.mapper.getObjectAttr(self.attr)
        if x is None:
            x = 0
        self.widget.setText(str(x))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        t = self.widget.text()
        x = 0 if t in [None, ""] else int(t)
        self.mapper.setObjectAttr(self.attr,x)

class BlankFloatValidator(QtGui.QDoubleValidator):
    def validate(self,input,pos):
        if input == "":
            return QtGui.QValidator.Acceptable, input, pos
        else:
            return QtGui.QDoubleValidator.validate(self,input,pos)

class FloatingPointYoke(LineYoke):
    def Factory(self):
        self.widget = QtGui.QLineEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.setValidator(BlankFloatValidator(self.widget))
        self.widget.editingFinished.connect(self.Save)
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def Bind(self):
        x = self.mapper.getObjectAttr(self.attr)
        if x is None:
            x = 0.
        self.widget.setText(str(x))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        t = self.widget.text()
        x = decimal.Decimal('0') if t in [None,""] else decimal.Decimal(t)
        self.mapper.setObjectAttr(self.attr,x)

class TextYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QtGui.QTextEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.setAcceptRichText(False)
        self.widget.setTabChangesFocus(True)
        # TODO:  figure out which signals to hook to Save
        # Qt QDataWidgetMapper uses installEventFilter
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def Bind(self):
        if isinstance(self.widget, QtGui.QLineEdit) or isinstance(self.widget, QtGui.QTextEdit):
            self.widget.setText(self.mapper.getObjectAttr(self.attr))
        elif isinstance(self.widget, QtGui.QPlainTextEdit):
            self.widget.setPlainText(self.mapper.getObjectAttr(self.attr))
        elif isinstance(self.widget, QtGui.QComboBox):
            self.widget.setEditText(self.mapper.getObjectAttr(self.attr))
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        if isinstance(self.widget, QtGui.QTextEdit) or isinstance(self.widget, QtGui.QPlainTextEdit):
            self.mapper.setObjectAttr(self.attr,self.widget.toPlainText())
        elif isinstance(self.widget, QtGui.QLineEdit):
            self.mapper.setObjectAttr(self.attr,self.widget.text())
        elif isinstance(self.widget, QtGui.QComboBox):
            self.mapper.setObjectAttr(self.attr,self.widget.currentText())
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

class FormattedYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QtGui.QTextEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.setTabChangesFocus(True)
        # TODO:  figure out which signals to hook to Save
        # Qt QDataWidgetMapper uses installEventFilter
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def Bind(self):
        if isinstance(self.widget, QtGui.QTextEdit):
            self.widget.setHtml(self.mapper.getObjectAttr(self.attr))
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        if isinstance(self.widget, QtGui.QTextEdit):
            self.mapper.setObjectAttr(self.attr,self.widget.toHtml())
        else:
            raise ValueError("The widget {0} is not supported by a {1}".format(self.widget, self.__class__.__name__))

class DateYoke(LineYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        from widgets import PBDateEdit
        self.widget = PBDateEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.editingFinished.connect(self.Save)
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def Bind(self):
        d = self.mapper.getObjectAttr(self.attr)
        if d is not None:
            d = toQType(d)
        self.widget.setDate(d)

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        d = self.widget.date
        if d is None:
            self.mapper.setObjectAttr(self.attr,None)
        else:
            self.mapper.setObjectAttr(self.attr,fromQType(d, datetime.date))

class TimeYoke(LineYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QTimeEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.editingFinished.connect(self.Save)
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        if isinstance(self.widget, QtGui.QLineEdit):
            self.widget.editingFinished.connect(self.Save)
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)

    def Bind(self):
        d = self.mapper.getObjectAttr(self.attr)
        if d is not None:
            d = toQType(d)
        else:
            d = QtCore.QTime(0,0)
        self.widget.setTime(d)

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        d = self.widget.time()
        if d is None:
            self.mapper.setObjectAttr(self.attr,None)
        else:
            self.mapper.setObjectAttr(self.attr,fromQType(d, datetime.time))

class BooleanYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QtGui.QCheckBox(ClassAttributeLabel(getattr(self.mapper.cls,self.attr)))
        self._baseAdoptWidget(self.widget)
        self.widget.toggled.connect(self.Save)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        if isinstance(self.widget, QtGui.QAbstractButton):
            self.widget.toggled.connect(self.Save)

    def Bind(self):
        self.widget.setChecked(toQType(self.mapper.getObjectAttr(self.attr), bool))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        self.mapper.setObjectAttr(self.attr,fromQType(self.widget.isChecked()))

    def LabelStyle(self):
        return InputYoke.Label.Internal

class BooleanRadioYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        raise NotImplementedError("somebody needs to make a UserAttr with descriptions for the radio buttons")
        #self.widget = QtGui.QCheckBox(ClassAttributeLabel(getattr(self.mapper.cls,self.attr)))
        #self.widget.toggled.connect(self.Save)
        #return self.widget

    def AdoptWidget(self, widget):
        """
        The widget parameter to :method:`WidgetAttributeMapper.bind` must be a 
        two-tuple of radio buttons when the bound yoke is a :class:`BooleanRadioYoke`. 
        The first widget is the radio button mapped to true and the second is 
        mapped to False.
        """
        assert len(widget) == 2, "I don't know what do to with this if it's not a list of length 2"
        self.true_radio = widget[0]
        self.false_radio = widget[1]
        
        if self.true_radio.group() is None:
            # it's unclear what an appropriate parent widget would be for the QButtonGroup
            # does it need one?
            # I think that using self.true_radio gets Qt in a minor funk because it might be a parent loop (It gets PySide in a funk for sure)
            self.group = QtGui.QButtonGroup(self.true_radio.parent())
            self.group.addButton(self.true_radio)
            self.group.addButton(self.false_radio)

        self.true_radio.clicked.connect(lambda clicked, value=True: self.radioChange(value, clicked))
        self.false_radio.clicked.connect(lambda clicked, value=False: self.radioChange(value, clicked))
        for w in [self.true_radio, self.false_radio]:
            self._baseAdoptWidget(w)

    def radioChange(self, value, clicked):
        if clicked:
            self.Save(value)

    def Bind(self):
        value = toQType(self.mapper.getObjectAttr(self.attr), bool)
        if value:
            self.true_radio.setChecked(True)
        else:
            self.false_radio.setChecked(True)

    def Save(self, value=None):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        if value is None:
            if self.true_radio.isChecked():
                value = True
            if self.false_radio.isChecked():
                value = False
        self.mapper.setObjectAttr(self.attr, value)

    def LabelStyle(self):
        return InputYoke.Label.Internal

def yokes_dict():
    d = {}
    d["date"] = DateYoke
    d["time"] = TimeYoke
    d["bool"] = BooleanYoke
    d["bool:options"] = BooleanRadioYoke
    d["int"] = IntegerYoke
    d["float"] = FloatingPointYoke
    d["text"] = TextYoke
    d["formatted"] = FormattedYoke
    d["line"] = LineYoke
    d["password"] = PasswordYoke

    from widgets import StreetAddressYoke
    d["address"] = StreetAddressYoke

    import foreign_key
    d["foreign_key"] = foreign_key.ForeignKeyEditYoke
    d["foreign_key_edit"] = foreign_key.ForeignKeyEditYoke
    d["foreign_key_combo"] = foreign_key.ForeignKeyComboYoke

    # TODO:  refine this ad-hoc means and provide sensible example
    if QtCore.QCoreApplication.instance() and hasattr(QtCore.QCoreApplication.instance(),"yokes"):
        d.update(QtCore.QCoreApplication.instance().yokes)

    return d

def addGlobalYoke(yoke, yokeFactory):
    """
    Add the yoke and yokeFactory to the global list of yokes.  Currently this 
    relies on the singleton QtCore.QCoreApplication.instance() by adding the 
    yoke to a dictionary set up in the yokes attribute of this class.  This 
    implementation detail has user ramifications, but is not considered an 
    ideal implementation.
    
    :param yoke:  the name of the yoke
    :param yokeFactory:  a callable taking a mapper class and an attribute
    """
    if not QtCore.QCoreApplication.instance():
        raise RuntimeError("The addGlobalYoke function relies on having a QtCore.QCoreApplication instance available")
    
    if not hasattr(QtCore.QCoreApplication.instance(),"yokes"):
        QtCore.QCoreApplication.instance().yokes = {}

    QtCore.QCoreApplication.instance().yokes[yoke] = yokeFactory

def YokeFactory(cls, attr, mapper, parent=None, index=None):
    """
    Create a yoke for the attribute `attr` in class `cls`.  Each sqlalchemy 
    column attribute and UserAttr can have a yoke specified for it in the 
    following ways:
      - specify a named yoke in a sqlalchemy column by including an info 
        parameter to the Column constructor.  For example::
        
            Column('name', Text, info={"yoke": "line"})

      - specify a named yoke in a UserAttr derived class by overriding the 
        yoke_specifier method and returning the name of the yoke

    The list of yoke names is extensible and individual yokes can be replaced by 
    calling :func:`addGlobalYoke`.
    """
    userattr = getattr(cls,attr)
    spec = ClassAttributeYoke(userattr)
    yoke_class = yokes_dict()[spec]
    return yoke_class(mapper,attr)



# -*- coding: utf-8 -*-
"""
We create models, attach to data and see if the events all propogate.
"""

from qtalchemy import UserAttr, ClassTableModel, ModelObject, UseCachedValue, instanceEvent
from PyQt4 import QtCore, QtGui

count = 0

class PseudoModel(object):
    def __init__(self,obj=None):
        self.obj = None
        if obj is not None:
            self.connect_instance(obj)

    def connect_instance(self,obj):
        self.obj = obj
        instanceEvent(self.obj, "set", "")(self.row_object_change)

    def row_object_change(self, obj, attr, oldvalue):
        assert obj is self.obj
        global count
        if attr == '__primality':
            count += 1

class IntegerProperties(ModelObject):
    def __init__(self,specimen=0):
        self.specimen = specimen
    
    prime = UserAttr(bool,'Prime',storage='__primality')

    @prime.on_get
    def is_specimen_prime(self):
        try:
            if self.specimen == self.__last_computed_prime:
                raise UseCachedValue()
        except AttributeError as e:
            pass # first run will except on __last_computed_prime

        result = False
        if self.specimen >= 2:
            result = True
            for i in xrange(2, self.specimen):
                if i * i > self.specimen:
                    break
                if (self.specimen % i) == 0:
                    result = False
                    break
        self.__last_computed_prime = self.specimen
        return result

def test_set_counter():
    global count

    assert count == 0, 'should start at 0'

    x = IntegerProperties(54)
    hold_reference = PseudoModel(x)

    assert count == 0, 'should start at 0'

    if x.prime == True:
        raise RuntimeError( '{0} is not prime'.format(x.specimen) )

    assert count == 1, 'checked once (count={0})'.format(count)

    if x.prime == True:
        raise RuntimeError( '{0} is not prime'.format(x.specimen) )

    assert count == 1, 'checked twice, but redundant (count={0})'.format(count)

    x.specimen = 97

    if x.prime == False:
        raise RuntimeError( '{0} is prime'.format(x.specimen) )

    assert count == 2, 'new number (count={0})'.format(count)

    if x.prime == False:
        raise RuntimeError( '{0} is prime'.format(x.specimen) )

    assert count == 2, 'new number; 2nd check (count={0})'.format(count)



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

import re
import decimal
import datetime

def default_for_type(type_):
    if type_ == str:
        return ""
    elif type_ == int:
        return 0
    elif type_ == datetime.date:
        return datetime.date(datetime.date.today().year,1,1)
    return type_()

def hasextendedattr(obj,attr):
    """
    >>> class Person:
    ...     name='Joe'
    ... 
    >>> class Group:
    ...     person1 = Person()
    ...     person2 = Person()
    ...
    >>> x = Group()
    >>> hasextendedattr(x,'person1.name')
    True
    >>> hasextendedattr(x,'person1')
    True
    >>> hasextendedattr(x,'person1.piggy')
    False
    """
    attrs = attr.split('.')
    o = obj
    #print attrs
    for a in attrs[:-1]:
        if hasattr(o,a):
            o = getattr(o,a)
        else:
            return False
    return hasattr(o,attrs[-1])

def getextendedattr(obj,attr):
    """
    >>> class Person:
    ...     name='Joe'
    ... 
    >>> class Group:
    ...     person1 = Person()
    ...     person2 = Person()
    ...
    >>> x = Group()
    >>> getextendedattr(x,'person1.name')
    'Joe'
    >>> getextendedattr(x,'person3.asdf')
    Traceback (most recent call last):
    ...
    AttributeError: Group instance has no attribute 'person3'
    >>> getextendedattr(x,'person2.asdf')
    Traceback (most recent call last):
    ...
    AttributeError: Person instance has no attribute 'asdf'
    """
    attrs = attr.split('.')
    o = obj
    for a in attrs:
        o = getattr(o,a)
    return o

def setextendedattr(obj,attr,value):
    """
    >>> class Person:
    ...     name='Joe'
    ... 
    >>> class Group:
    ...     person1 = Person()
    ...     person2 = Person()
    ...
    >>> x = Group()
    >>> setextendedattr(x,'person1.name','Sam')
    >>> x.person1.name
    'Sam'
    >>> x.person2.name
    'Joe'
    """
    attrs = attr.split('.')
    o = obj
    for a in attrs[:-1]:
        o = getattr(o,a)
    setattr(o,attrs[-1],value)

class UseCachedValue(Exception):
    """
    Raise this exception from UserAttr.on_get handlers to avoid returning a 
    new value.  This is useful when writing a handler for a calculated field 
    and the cached value is correct.  In these cases, we want to avoid 
    returning the cached value as newly computed because that will incur data 
    model updates which are not needed.
    """
    pass

class AttributeInstrumentation(object):
    """
    This is an internal class for the UserAttr objects.  It has flags for 
    recursion control in the getters and setters.
    """
    def __init__(self):
        self.getting = 0
        self.setting = 0

class UserAttr(property):
    """
    The UserAttr type provides python object attribute type validation.
    
    The UserAttr class is designed to be used in conjunction with
    :class:`qtalchemy.ModelObject`.  

    Currently the on_get event semantics is inconsistent with the use of the
    ModelObject.Events event model.  That is the UserAttr.on_get decorator is
    on the attribute itself, but the ModelObject.Events.add operator is based
    on an instance of ModelObject.Events.  The advantage of the latter is that
    it works consistently for both UserAttr and sqlalchemy column objects.

    EXAMPLE::
        >>> class Person(ModelObject):
        ...     name = UserAttr(str,"Person Name")
        ...     age = UserAttr(int,"Age")
        ...     married = UserAttr(bool,"Married")
        >>> 
        >>> fred = Person()
        >>> fred.name="Fred Jackson"
        >>> fred.age="1234"
        >>> fred.married = 4
        >>> (fred.name, fred.age, fred.married)
        ('Fred Jackson', 1234, True)
        >>> type(fred.age)
        <type 'int'>
        >>> type(fred.married)
        <type 'bool'>

        >>> jane = Person()
        >>> jane.age
        0
        >>> jane.name
        ''
        >>> jane.name=None
        >>> jane.name
        ''
        >>> jane.age=None
        >>> jane.age
        0
    """
    def __init__(self,atype,label,storage=None,readonly=False,whats_this=None):
        """
        The UserAttr property extends python 'property' class to provide a type and label and default storage.
        
        :param atype: python type of the UserAttr
        :param label: label for display to application user
        :param storage: attribute which stores this in the object instances 
         - if not specified, it is derived by turning label into a python safe name with leading underscores, etc.
         - storage can be specified as a dot qualified multi-identifier and the components will be resolved iteratively
        :param readonly: display this property as read-only in the interface
        """
        property.__init__(self,self.fget,self.fset)
        self.atype = atype
        if storage is not None:
            self.storage = storage
        else:
            self.storage = "_UserAttr_" + re.sub("[^a-zA-Z0-9_]","_",label)
        self.label = label
        self.readonly = readonly
        self.whats_this = whats_this
        self.__listeners = []
        self.__on_first_get = None
        self.__on_get = None

    @property
    def base_type(self):
        if hasattr(self.atype, "base_type"):
            return self.atype.base_type
        else:
            return self.atype

    def _instrumentation(self,row):
        attrName = "_instrumentation_"+self.storage.replace('.','_')
        if not hasattr(row,attrName):
            setattr(row,attrName,AttributeInstrumentation())
        return getattr(row,attrName)

    def fget(self,row):
        inst = self._instrumentation(row)
        inst.getting+=1
        if self.__on_get is not None:
            try:
                self.fset(row,self.__on_get(row))
            except UseCachedValue as e:
                pass
        elif not hasextendedattr(row, self.storage) and self.__on_first_get is not None:
            try:
                self.fset(row,self.__on_first_get(row))
            except UseCachedValue as e:
                pass
        inst.getting-=1
        return getextendedattr(row,self.storage) if hasextendedattr(row, self.storage) else default_for_type(self.base_type)

    def on_first_get(self,func):
        self.__on_first_get = func
        return func

    def on_get(self,func):
        self.__on_get = func
        return func

    def on_change(self,func):
        self.__listeners.append(func)
        return func

    def fset(self,row,value):
        inst = self._instrumentation(row)
        assert 0 == inst.setting, "Recursive set of {0}".format(self.storage)
        inst.setting+=1
        try:
            # if the underlying type is a 'type' and the value matches that type, we can skip directly to the assignment.
            if not isinstance(self.atype,type) or not isinstance(value,self.atype):
                if value is None:
                    value = self.atype()
                elif value in [""] and self.atype in (float,int,decimal.Decimal):                
                    value = self.atype("0")
                else:
                    value = self.atype(value)

            setextendedattr(row,self.storage,value)

            # notify 
            if 0 == self._instrumentation(row).getting:
                # Note that ModelObject.__setattr__ accounts for both SA 
                # columns and UserAttr during attribute setting
                for l in self.__listeners:
                    l(row)
        finally:
            inst.setting-=1

    def session(self, row):
        try:
            return row.session()
        except AttributeError as e:
            return None

    def session_maker(self, row):
        return self.session(row).__class__

class Nullable:
    """
    A type extender for use with UserAttr.  Wrapping a type with Nullable allows valid values for that type or None.
    
    EXAMPLE::
        >>> class Thing(object):
        ...     date = UserAttr(Nullable(datetime.date),"Optional Date")
        >>> 
        >>> t = Thing()
        >>> t.date = None          # succeeds and sets the date to None
        >>> t.date is None
        True
        >>> t.date = 12            # This will fail, datetime.date doesn't coerce an int.
        Traceback (most recent call last):
        ...
        TypeError: Required argument 'month' (pos 2) not found
        >>> t.date = datetime.date(2010,1,9) # my birthday
        >>> t.date
        datetime.date(2010, 1, 9)
    """
    def __init__(self,atype):
        self.atype = atype

    def __call__(self,v=None):
        if v is None:
            return v
        elif not isinstance(v,self.atype):
            return self.atype(v)
        else:
            return v

    @property
    def base_type(self):
        return self.atype

    def yoke_specifier(self):
        type_yoke_specifier(self.atype)

class AttrNumeric:
    """
    This class wraps decimal.Decimal with construction semantics compatible with the UserAttr class.
    
    EXAMPLE::
        >>> AttrNumeric(4)(12)
        Decimal('12.0000')
        >>> AttrNumeric(1)(23.352)
        Decimal('23.4')
        >>> AttrNumeric(1)("-3.9")
        Decimal('-3.9')

        
        >>> class BankAccount(object):
        ...     account_no = UserAttr(str,"Account Number")
        ...     open_date = UserAttr(datetime.date,"Opened")
        ...     balance = UserAttr(AttrNumeric(2),"Balance")
        ... 
        >>> b = BankAccount()
        >>> b.balance
        Decimal('0.00')
        >>> b.balance = 12.5352
        >>> b.balance
        Decimal('12.54')
        >>> b.balance = 3.235
        >>> b.balance
        Decimal('3.24')
        >>> b.balance = 12
        >>> b.balance
        Decimal('12.00')
        >>> b.balance = decimal.Decimal(100)/3
        >>> b.balance
        Decimal('33.33')
        >>> b.balance = decimal.Decimal(7)/200
        >>> b.balance
        Decimal('0.04')
    """
    def __init__(self, prec):
        assert prec>=0, "numeric precision must be at least zero"
        self.prec=prec

    def __call__(self, value=None):
        if value in ["", None]:
            return decimal.Decimal()._rescale(-self.prec,'ROUND_HALF_UP')
        elif type(value) in [float]:
            # take higher precision when converting to a string
            v = decimal.Decimal("%.*f" % (self.prec+2, value))
            #print value, v
            return v._rescale(-self.prec,'ROUND_HALF_UP')
        else:
            v = decimal.Decimal(value)
            return v._rescale(-self.prec,'ROUND_HALF_UP')

    def yoke_specifier(self):
        if self.prec==0:
            return "int"
        else:
            return "float"



# -*- coding: utf-8 -*-
"""
We create models, attach to data and see if the events all propogate.
"""

from qtalchemy import UserAttr, AttrNumeric, ClassTableModel, ModelObject
from PyQt4 import QtCore, QtGui

class ItemTest(ModelObject):
    def __init__(self,item=""):
        self.item = item
    
    item = UserAttr(str,"item")
    price = UserAttr(AttrNumeric(2),"price")

def create_model_test():
    item_names = ["wrench", "hammer"]
    items = [ItemTest(item=b) for b in item_names]
    
    model = ClassTableModel(ItemTest,"item,price".split(','),readonly=False)
    model.reset_content_from_list(items)

    model.setData(model.index(1,1,None),3.235,QtCore.Qt.EditRole)

    model.setData(model.index(0,1,None),4,QtCore.Qt.EditRole)
    
    assert items[0].price == 4
    assert items[1].price*100 == 324



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
This module brings together a number of crucial elements of the qtalchemy library.  It 
depends on the UserAttr module and builds in so widget knowledge of the UserAttr (or particular 
for foreign keys).  This requires tying in sqlalchemy data sources and makes some assumptions 
about the qtalchemy approach to building a dialog.
"""

from PyQtHelpers import *
from sqlalchemy.orm import mapper, create_session, relation, object_session, Query
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy_helper import *
import sqlalchemy
import datetime
import decimal
from user_attr import UserAttr, AttrNumeric

if sqlalchemy.__version__.split('.') < ['0', '6', '0']:
    sqlalchemy_to_python_type = [(satype,pytype) for pytype,satype in sqlalchemy.types.type_map.iteritems()]
elif sqlalchemy.__version__.split('.') < ['0', '7', '0']:
    sqlalchemy_to_python_type = [(satype.__class__,pytype) for pytype,satype in sqlalchemy.types.type_map.iteritems()]
else:
    # TODO:  if we weren't convinced before that type_map was a secret, now we should be :)
    sqlalchemy_to_python_type = [(satype.__class__,pytype) for pytype,satype in sqlalchemy.types._type_map.iteritems()]
sqlalchemy_to_python_type.append((sqlalchemy.types.String,str))

def BoundPropertyAttribute(classMember, attribute):
    """
    Abstracts out property determination code between a UserAttr and SA column.

    Attributes like the label, yoke and what's this string are kept in the 
    SA info dictionary for a column.  They are kept directly as members of the 
    descriptor class for a UserAttr.
    """

    # classMember appears like a SA column
    if hasattr(classMember,"label") and callable(classMember.label):
        col=classMember.property.columns[0]
        if hasattr(col,"info") and col.info.has_key(attribute):
            return col.info[attribute]

    try:
        # cls has an attribute col that looks like a UserAttr
        return getattr(classMember, attribute)
    except AttributeError:
        pass

    return None

def ClassAttributeWhatsThis(classMember):
    return BoundPropertyAttribute(classMember, "whats_this")

def UserLabel(identifier):
    """
    Convert a python or SQL identifier name to a user viewable label.  Essentially this makes word breaks 
    at '_' characters and uses initial caps for words.
    
    >>> UserLabel("balance_sheet")
    'Balance Sheet'
    >>> UserLabel("account")
    'Account'
    """
    return identifier.replace('_', ' ').title()

def ClassAttributeLabel(classMember):
    """
    Abstracts out the label determination code between a UserAttr and SA column.
    
    See also :ClassAttributeType
    >>> class X:
    ...     name=UserAttr(str,"First Name")
    >>> ClassAttributeLabel(X.name)
    'First Name'

    >>> from sqlalchemy import Table, Column, Integer, String, Text, Date, Boolean, Numeric, MetaData
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> metadata = MetaData()
    >>> Base = declarative_base(metadata=metadata)
    >>> class Options(Base):
    ...     __table__ = Table('options', metadata,
    ...             Column('id', Integer, primary_key=True),
    ...             Column('data_name', String(50)),
    ...             Column('company', String(50), info={"label": "Company Name"}))
    >>> ClassAttributeLabel(Options.data_name)
    'Data Name'
    >>> ClassAttributeLabel(Options.company)
    'Company Name'
    """

    # classMember appears like a SA column
    if hasattr(classMember,"label") and callable(classMember.label):
        col=classMember.property.columns[0]
        if hasattr(col,"info") and col.info.has_key("label"):
            return col.info["label"]
        else:
            return UserLabel(col.name)

    try:
        # cls has an attribute col that looks like a UserAttr
        return classMember.label
    except AttributeError:
        pass

    return None

def attrLabel(cls, col):
    """
    See also attrType and ClassAttributeLabel

    This helper function returns a label from class and attribute pair.  It handles the following cases:
     - cls is a new-style python class and col is a string representing the attribute name of a UserAttr
     - cls is a new-style python class and col is a string representing the attribute name of a SqlAlchemy column
     - cls is None and col is a ModelColumn object
    
    >>> m=ModelColumn("name",str,"First Name")
    >>> attrLabel(None,m)
    'First Name'
    """
    try:
        # col looks like a ModelColumn
        return col.label
    except AttributeError:
        assert cls is not None, "If 'col' is not (like) a ModelColumn, then cls must be specified."

    return ClassAttributeLabel(getattr(cls,col))

def ClassAttributeYoke(classMember):
    """
    :param classMember: The attribute in the class (not the instance)
    
    This function
    
    >>> class X:
    ...     name=UserAttr(str,"First Name")
    ...     age=UserAttr(AttrNumeric(3),"Precise Age")
    >>> ClassAttributeYoke(X.name)
    'line'
    >>> ClassAttributeYoke(X.age)
    'float'

    >>> from sqlalchemy import Table, Column, Integer, String, Text, Date, Time, Boolean, Numeric, MetaData
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> metadata = MetaData()
    >>> Base = declarative_base(metadata=metadata)
    >>> class Books(Base):
    ...     __table__ = Table('books', metadata,
    ...             Column('id', Integer, primary_key=True),
    ...             Column('author', String(50)),
    ...             Column('release', Date()),
    ...             Column('release_time', Time()),
    ...             Column('copies', Integer),
    ...             Column('juvenile', Boolean),
    ...             Column('review', Text),
    ...             Column('price', Numeric(precision=10,scale=2,asdecimal=True)),
    ...             Column('title', String(50), info={'yoke': 'text'}))
    >>> ClassAttributeYoke(Books.author)
    'line'
    >>> ClassAttributeYoke(Books.release)
    'date'
    >>> ClassAttributeYoke(Books.release_time)
    'time'
    >>> ClassAttributeYoke(Books.copies)
    'int'
    >>> ClassAttributeYoke(Books.price)
    'float'
    >>> ClassAttributeYoke(Books.title)
    'text'
    >>> ClassAttributeYoke(Books.review)
    'text'
    >>> ClassAttributeYoke(Books.juvenile)
    'bool'
    """

    type_=None

    try:
        # unspecified UserAttr class
        type_ = classMember.base_type
        return classMember.yoke_specifier()
    except AttributeError, e:
        pass

    if type_ is not None:
        try:
            # specified UserAttr class?
            return type_.yoke_specifier()
        except AttributeError, e:
            pass

    # classMember appears like a SA column
    if type_ is None and hasattr(classMember,"label") and callable(classMember.label):
        col=classMember.property.columns[0]
        if hasattr(col,"info") and col.info.has_key("yoke"):
            return col.info["yoke"]
        else:
            if isinstance(classMember.property.columns[0].type, sqlalchemy.types.Text):
                return "text"
            for satype,pytype in sqlalchemy_to_python_type:
                if isinstance(classMember.property.columns[0].type,satype):
                    type_ = pytype

    if type_ == datetime.date:
        return "date"
    elif type_ == datetime.time:
        return "time"
    elif type_ == bool:
        return "bool"
    elif type_ == int:
        return "int"
    elif type_ in (decimal.Decimal,float):
        return "float"
    else:
        return "line"

def attrYoke(cls, col):
    """
    See also attrType and ClassAttributeLabel

    This helper function returns a label from class and attribute pair.  It handles the following cases:
     - cls is a new-style python class and col is a string representing the attribute name of a UserAttr
     - cls is a new-style python class and col is a string representing the attribute name of a SqlAlchemy column
     - cls is None and col is a ModelColumn object
    
    >>> m=ModelColumn("name",str,"First Name")
    >>> attrYoke(None,m)
    'line'
    """
    if cls is not None:
        return ClassAttributeYoke(getattr(cls,col.attr))

    try:
        # col looks like a ModelColumn
        type_ = col.type_

        if type_ == datetime.date:
            return "date"
        elif type_ == bool:
            return "bool"
        elif type_ == int:
            return "int"
        elif type_ in (decimal.Decimal,float):
            return "float"
        else:
            return "line"
    except AttributeError:
        assert cls is not None, "If 'col' is not (like) a ModelColumn, then cls must be specified."

    #return ClassAttributeYoke(getattr(cls,col))

def ClassAttributeType(classMember):
    """
    :param classMember: The attribute in the class (not the instance)
    
    This function
    
    >>> class X:
    ...     name=UserAttr(str,"First Name")
    >>> ClassAttributeType(X.name)
    <type 'str'>

    >>> from sqlalchemy import Table, Column, Integer, String, Text, Date, Boolean, Numeric, MetaData
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> metadata = MetaData()
    >>> Base = declarative_base(metadata=metadata)
    >>> class Books(Base):
    ...     __table__ = Table('books', metadata,
    ...             Column('id', Integer, primary_key=True),
    ...             Column('author', String(50)),
    ...             Column('release', Date()),
    ...             Column('copies', Integer),
    ...             Column('juvenile', Boolean),
    ...             Column('price', Numeric(precision=10,scale=2,asdecimal=True)),
    ...             Column('title', String(50)))
    >>> ClassAttributeType(Books.author)
    <type 'str'>
    >>> ClassAttributeType(Books.release)
    <type 'datetime.date'>
    >>> ClassAttributeType(Books.copies)
    <type 'int'>
    >>> # platform confusion on the next one
    >>> ClassAttributeType(Books.price) in (decimal.Decimal, float)
    True
    >>> ClassAttributeType(Books.juvenile)
    <type 'bool'>
    """
    try:
        # cls has an attribute col that looks like a UserAttr
        return classMember.base_type
    except AttributeError:
        pass

    # cls is sqlalchemy mapped object with column col
    try:
        for satype,pytype in sqlalchemy_to_python_type:
            if isinstance(classMember.property.columns[0].type,satype):
                return pytype
    except AttributeError:
        pass

    return None # we don't know

def attrType(cls, col):
    """
    See also attrLabel

    This helper function returns a type from class and attribute pair.  It handles the following cases:
     - cls is a new-style python class and col is a string representing the attribute name of a UserAttr
     - cls is a new-style python class and col is a string representing the attribute name of a SqlAlchemy column
     - cls is None and col is a ModelColumn object
    
    >>> m=ModelColumn("name",str,"First Name")
    >>> attrType(None,m)
    <type 'str'>
    """
    try:
        # col looks like a ModelColumn
        return col.type_
    except AttributeError:
        assert cls is not None, "If 'col' is not (like) a ModelColumn, then cls must be specified."

    return ClassAttributeType(getattr(cls,col))

def attrWidth(cls, col):
    """
    Returns the width of a ModelColumn or SA column.
    """
    try:
        return col.width
    except AttributeError:
        pass
    return None

def attrReadonly(cls, col):
    try:
        return col.readonly
    except AttributeError:
        pass
    try:
        return getattr(cls,col).readonly
    except AttributeError:
        pass
    return False


class AlchemyModelDelegate(QtGui.QStyledItemDelegate):
    def setEditorData(self,editor,index):
        try:
            cls = index.model().cls
            if cls:
                c = index.model().columns[index.column()]
                userattr = getattr(cls,c.attr if hasattr(c,"attr") else c)
                if userattr is not None and hasattr(userattr,"WidgetBind"):
                    userattr.WidgetBind(editor,index.model().row_object_by_index(index))
        except AttributeError, e:
            pass

        # http://bugreports.qt.nokia.com/browse/QTBUG-428 - combo boxes don't play nice with qdatawidget mapper 
        if isinstance(editor,QtGui.QComboBox):
            text = index.data()
            if text == None:
                text = ""
            if editor.isEditable():
                editor.setEditText(text)
            else:
                editor.setCurrentIndex(editor.findText(text))
        else:
            QtGui.QStyledItemDelegate.setEditorData(self,editor,index)

    def setModelData(self,editor,model,index):
        if isinstance(editor,QtGui.QComboBox):
            model.setData(index,editor.currentText())
        else:
            QtGui.QStyledItemDelegate.setModelData(self,editor,model,index)

    def createEditor(self,parent,option,index):
        try:
            cls = index.model().cls
            if cls:
                c = index.model().columns[index.column()]
                userattr = getattr(cls,c.attr if hasattr(c,"attr") else c)
                if userattr is not None and hasattr(userattr,"WidgetFactory"):
                    return userattr.WidgetFactory(parent, index)
        except AttributeError, e:
            pass
        return QtGui.QStyledItemDelegate.createEditor(self,parent,option,index)

def AttrWidgetFactory(userattr,parent=None, index=None):
    if hasattr(userattr,"WidgetFactory"):
        return userattr.WidgetFactory(parent, index)
    myAttrType = ClassAttributeType(userattr)
    if myAttrType == datetime.date:
        from qtalchemy.widgets import PBDateEdit
        w = PBDateEdit(parent)
        #w = QtGui.QDateEdit(parent)
        #w.setCalendarPopup(True)
        return w
    elif myAttrType == bool:
        return QtGui.QCheckBox(ClassAttributeLabel(userattr), parent)
    elif myAttrType == int:
        w = QtGui.QLineEdit(parent)
        w.setValidator(QtGui.QIntValidator())
        return w
    elif myAttrType in (decimal.Decimal,float):
        w = QtGui.QLineEdit(parent)
        w.setValidator(QtGui.QDoubleValidator(parent))
        return w
    else:
        return QtGui.QLineEdit(parent)


class ModelColumn:
    def __init__(self, attr, type_, label=None, readonly=False, width=None):
        """
        :param attr:  attribute name
        :param label:  label to display in the list header
        :param readonly:  true if the column should not be editable even if the view is
        :param width:  width in pixels (?) or relative width to other columns
        """
        self.attr = attr
        if label is None:
            self.label = attr.title().replace("_"," ")
        else:
            self.label = label
        self.type_ = type_
        self.readonly = readonly
        self.width = width


class ObjectRepresent(object):
    def __init__(self, idattr, idListMimeType=None):
        """
        Instances of this class can be used for an objectConverter for a 
        PBTableModel.  Such instances take care of resolving the primary key 
        for passing to a CommandEvent.  They also fill a QMimeData with a list 
        of the primary keys in a set of model indexes.
        
        A model exposing mime data in this way can be used as a source of items 
        draggable with actions on drop defined by a function decorated with 
        :func:`CommandMenu.dropCommand`.
        
        :param idattr:  The attribute which holds the primary key of this object
        :param idListMimeType:  The mime type for a list of id's.
        """
        self.idattr = idattr
        self.idListMimeType = idListMimeType
        
    def __call__(self, obj):
        return getattr(obj, self.idattr)

    def mime(self, qmime, model, indexes):
        if self.idListMimeType is not None:
            ids = set([self(x.internalPointer()) for x in indexes])
            qmime.setData(self.idListMimeType, '\n'.join([str(x) for x in ids]))


class SACollectionManagerForModel:
    """
    Provide semi-hookable functions from PBTableModel for adding and deleting
    rows.

    The PBTableModel includes an additional blank (or multiple blank) row at the
    bottom of the list of existing lines. This line is not the SQLAlchemy
    session (or other collection class), but it is there for a user convenience
    to add a new line.
    """
    def __init__(self,klass,collection=None,session=None,fixed_rows=False):
        self.klass = klass
        self.fixed_rows = fixed_rows
        self.reset(collection, session)

    def reset(self,collection,session):
        self.session = session
        self.collection = collection

        # TODO: consider getting the session via:
        #self.session = object_session(self.collection._sa_adapter.owner_state.obj())

    def add_flipper_row(self):
        """
        Create a new row, but do not add it to the sqlalchemy session.
        
        A 'flipper row' is a row typically at the bottom of the list which is
        not yet in the (sql-alchemy) collection, but appears visually.
        """
        r = self.klass()
        # The object is not added to the session, but we like to pretend that it is.
        if self.session is not None:
            self.session.npadd(r)
        return r

    def join_flipper(self, flipper):
        """
        Add a pre-existing object to the collection
        """
        self.collection += [flipper]

    def add_row(self):
        """
        Make a new row object, add it to the collection and return it.
        
        Note that in the usual (?) PBTableModel use case, a new row passes through 
        a flipper row stage.  The add_row method can be viewed as the sum of 
        add_flipper_row and join_flipper.
        """
        r = self.klass()
        # implicitly added to the session when added to the collection
        self.collection += [r]
        return r

    def delete_row(self, row):
        """
        Remove the row object from the collection.  Return True if successful, False otherwise.
        """
        try:
            self.collection.remove(row)
            # we must explicitly delete the row from the session to remove it from the database
            session = object_session(self.collection._sa_adapter.owner_state.obj())
            if session is not None:
                session.delete(row)
            return True
        except InvalidRequestError as e:
            session.expunge(row)
            return True
        except Exception as e:
            return False

def ClassTableModel(cls, attributes, readonly=True, fixed_rows=False, row_manager=None, objectConverter=None, tree_structure=None):
    """
    Create a :class:`PBTableModel` exposing a list of objects of a particular python class.
    
    Example:
        >>> from sqlalchemy import Table, Column, Integer, String, Text, Date, Boolean, Numeric, MetaData, ForeignKey, create_engine, select, join
        >>> from sqlalchemy.orm import Query
        >>> from sqlalchemy.ext.declarative import declarative_base
        >>> 
        >>> from qtalchemy import QueryTableModel, PBSessionMaker, ModelObject
        >>> from PyQt4 import QtCore, QtGui
        >>> 
        >>> metadata = MetaData()
        >>> Base = declarative_base(metadata=metadata,cls=ModelObject)
        >>> 
        >>> class Person(Base):
        ...     __table__ = Table('person',metadata,
        ...         Column('id', Integer, primary_key=True),
        ...         Column('name', String(50)))
        ...     nicks = relation("Nicks", backref="Person")
        ... 
        >>> class Nicks(Base):
        ...     __table__ = Table('nicks',metadata,
        ...         Column('id', Integer, primary_key=True),
        ...         Column('person_id', Integer, ForeignKey('person.id')),
        ...         Column('nick', String(50)))
        ...     def __repr__(self):
        ...         return "Nick:  %s" % (self.nick,)
        ... 
        >>> engine = create_engine("sqlite://")
        >>> metadata.bind = engine
        >>> metadata.create_all()
        >>> Session = PBSessionMaker(bind=engine)
        >>> session = Session()
        >>> p=Person()
        >>> session.add(p)
        >>> p.nicks
        []
        >>> model=ClassTableModel(Nicks,["nick"],readonly=False)
        >>> model.reset_content_from_list(p.nicks,session)
        >>> model.rowCount(None)
        1
        >>> index = model.index(0,0,None)
        >>> index.internalPointer()
        Nick:  None
        >>> index.internalPointer().session() is not None
        True
        >>> model.data(index,QtCore.Qt.DisplayRole)
        ''
        >>> model.setData(index,"sam",QtCore.Qt.EditRole)
        True
        >>> p.nicks
        [Nick:  sam]
        >>> index = model.index(1,0,None)
        >>> model.setData(index,"jack",QtCore.Qt.EditRole)
        True
        >>> p.nicks
        [Nick:  sam, Nick:  jack]
    """
    def makeModelColumn(cls, attr):
        ua = getattr(cls, attr)
        return ModelColumn(attr, ClassAttributeType(ua), ClassAttributeLabel(ua), readonly=attrReadonly(cls, attr))
    cols = [makeModelColumn(cls, a) for a in attributes]
    return PBTableModel(cols, cls=cls, readonly=readonly, fixed_rows=fixed_rows, row_manager=row_manager, objectConverter=objectConverter, tree_structure=tree_structure)


def QueryClassTableModel(cls, attributes, query, ssrc, objectConverter=None, tree_structure=None):
    """
    Construct a :class:`PBTableModel` from a sqlalchemy query returning a mapped 
    class.
    
    :param cls:  The mapped class which is returned
    :param attributes:  attributes to display
    :param query: the query object
    :param ssrc:  the session maker
    :param objectConverter: function taking an instance of cls and returning the primary key for use
    """
    def makeModelColumn(cls, attr):
        ua = getattr(cls, attr)
        return ModelColumn(attr, ClassAttributeType(ua), ClassAttributeLabel(ua), readonly=attrReadonly(cls, attr))
    cols = [makeModelColumn(cls, a) for a in attributes]
    return PBTableModel(cols, query=query, cls=cls, ssrc=ssrc, objectConverter=objectConverter, tree_structure=tree_structure)

def QueryTableModel(q, cols=None, include_uuid=False, ssrc=None, objectConverter=None, tree_structure=None):
    """
    This helper function constructs an :class:`PBTableModel` from a sqlalchemy query.  It can
    auto-detect the columns and types or take a list of column names from the query.
    
    :param q: A sqlalchemy.orm.query.Query object
    :param cols: an optional list of columns (if not specified, it displays all subject to include_uuid)
    :param include_uuid: optional parameter defaulting to False.  If include_uuid is true, then the generated
        table will include the uuids.  Generally, we only desire to have the uuid's around for use in the handlers.
    """
    model_cols = []
    uuid_cols = []
    sq = q.subquery() # need this to get column types
    for col in sq.columns:
        if col.name.startswith("_hidden_"):
            pass
        elif isinstance(col.type,UUID):
            if include_uuid:
                model_cols.append(ModelColumn(UserLabel(col.name),str))
            uuid_cols.append(col)
        else:
            match = False
            for satype,pytype in sqlalchemy_to_python_type:
                if isinstance(col.type,satype):
                    model_cols.append(ModelColumn(col.name,pytype))
                    match=True
                    break
            assert match, "The sqlalchemy type %s was not known for mapping to a python type." % col.type

    if objectConverter is None and len(uuid_cols)>0:
        # we take the first UUID as the key column.
        # It is a basic assumption of qtalchemy that everything has a UUID pk
        objectConverter = lambda x, col=uuid_cols[0]: getattr(x, col.name)

    if q.session is None:
        return PBTableModel(model_cols, query=q, ssrc=ssrc, objectConverter=objectConverter, tree_structure=tree_structure)
    else:
        return PBTableModel(model_cols, rows=q.all(), objectConverter=objectConverter, tree_structure=tree_structure)

class PBTableModel(QtCore.QAbstractItemModel):
    """
    A PyQt/PySide model for objects with :class:`UserAttr`.
    """
    def __init__(self, columns, cls=None, rows=None, query=None, ssrc=None, parent=None, readonly=True, row_manager=None, fixed_rows=False, objectConverter=None, tree_structure=None):
        """
        All parameters other than columns are optional.

        :param columns:  A list of columns.  These could be attribute names or ModelColumn objects.  Both methods can be used in the same model instance.
        :param query:  This could be SA Query object or a list of row objects
        :param cls:  The class of objects in the row list.
        :param ssrc:  A sqlalchemy sessionmaker.
        :param objectConverter: Callable taking row objects to unique keys in the application domain
        """
        QtCore.QAbstractItemModel.__init__(self,parent)
        self.columns = columns
        self.cls = cls

        self.objectConverter = (lambda x: x) if objectConverter is None else objectConverter

        self.query = query
        self.ssrc = ssrc

        self.row_list = rows
        self.readonly=readonly
        if row_manager is None:
            row_manager = SACollectionManagerForModel(self.cls, fixed_rows=fixed_rows)
        self.row_manager = row_manager
        
        # dropCallable is a mapping of mime-types to callable functions
        self.dropCallable = {}
        self.tree_structure = tree_structure

        self.setRoleNames(dict([(QtCore.Qt.UserRole+i,self.columns[i].attr) for i in range(len(columns))]))

        # making new rows and being readonly seems contradictory to me.
        assert not self.row_manager.fixed_rows or not self.readonly, "I'm not sure what a non-sessioned query means with an editable table."
        assert not self.row_manager.fixed_rows or self.query is None, "I'm not sure what a non-sessioned query means with an editable table."

        # The "flipper row" is the new row at the bottom of the list.
        self.flipper_rows = []
        if not self.readonly and not self.row_manager.fixed_rows and len(self.flipper_rows) == 0:
            self.flipper_rows.append(self.row_manager.add_flipper_row())

    def reset_content_from_list(self, rows, session=None):
        self.beginResetModel()
        self.row_list = rows

        # TODO:  Here is a dilemma:
        #  pull #1)  I'd like to keep requirements for classes with UserAttr and 
        #            hooking a model very slim
        #  pull #2)  I need some instrumentation to notify the PBTableModel
        #            of changes at the python object level
        #
        # Potential Fix:  Have the PBTableModel add an instrumentation object when 
        # needed.
        if self.cls is not None and issubclass(self.cls, ModelObject):
            for r in rows:
                instanceEvent(r, "set", "")(self.row_object_change)

        if self.row_manager is not None:
            self.row_manager.reset(self.row_list,session)
        self.flipper_rows = []
        if not self.readonly and not self.row_manager.fixed_rows and len(self.flipper_rows) == 0:
            self.flipper_rows.append(self.row_manager.add_flipper_row())
        self.endResetModel()

    def reset_content_from_session(self):
        """
        Reload the query from the session source 'ssrc' passed at object invocation.
        """
        if self.ssrc is None:
            return  # we don't know how to reload; make it a no-op for now
        
        q2 = self.query
        q2.session = self.ssrc()
        self.reset_content_from_list(q2.all())
        q2.session.close()
        q2.session = None

    def objParent(self, rowObj):
        if self.tree_structure is None:
            return None
        elif rowObj is None:
            return None
        else:
            return getattr(rowObj, self.tree_structure[0])

    def objChildren(self, rowObj):
        if self.tree_structure is None:
            return None
        else:
            return getattr(rowObj, self.tree_structure[1])

    def rowReverseIndex(self, row_obj):
        if self.objParent(row_obj) is not None:
            row_group = self.objChildren(self.objParent(row_obj))
            for i in range(len(row_group)):
                if row_group[i] == row_obj:
                    return self.createIndex(i, 0, row_obj)
        else:
            for i in range(self.rowCount(QtCore.QModelIndex())):
                row = self.row_object(i, None)
                if row is row_obj:
                    return self.createIndex(i, 0, row_obj)
        return QtCore.QModelIndex()

    def rowEmitChange(self, row_obj, col):
        """
        Emit change events for the row and column indexes indicated.
        
        :param row_obj: The object in the list of rows changed
        :param col: a column index or the string 'all' indicating 0 .. columnCount()-1
        """
        if col == "all":
            col1 = 0
            col2 = self.columnCount(None)-1
        else:
            col1 = col2 = col
        i = self.rowReverseIndex(row_obj)
        if i.isValid():
            self.dataChanged.emit(self.index(i.row(),col1,i.parent()),self.index(i.row(),col2,i.parent()))

    def row_object_change(self, row_obj, key, oldvalue):
        col_changed = -1
        for i in range(len(self.columns)):
            if self.columns[i].attr == key:
                col_changed = i
                break
        if col_changed == -1:
            return  # if the changing key is not a member of this model, then nothing to do here
        self.rowEmitChange(row_obj, col_changed)

    def row_object_by_index(self, index):
        return index.internalPointer()

    def row_object(self, row, parent):
        if parent is None:
            row_group = self.row_list
        else:
            row_group = self.objChildren(parent)

        if row_group is None or row < 0:
            return None

        if row < len(row_group):
            return row_group[row]
        else:
            return self.flipper_rows[row - len(row_group)]

    def index(self, row, column, parent):
        if parent is None:
            parent = QtCore.QModelIndex()
        return self.createIndex(row, column, self.row_object(row, parent.internalPointer()))

    def parent(self, child):
        obj = child.internalPointer()
        parent = self.objParent(obj)
        if parent is None:
            return QtCore.QModelIndex()
        else:
            return self.rowReverseIndex(parent)

    def columnCount(self,parent):
        #print "columnCount", len(self.columns)
        return len(self.columns)

    def rowCount(self, parent):
        if self.row_list is None:
            return 0

        #print "rowCount", len(self.row_list)
        if self.tree_structure is None or parent is None or not parent.isValid():
            if self.row_manager.fixed_rows:
                return len(self.row_list)
            else:
                return len(self.row_list) + len(self.flipper_rows)
        else:
            obj = parent.internalPointer()
            return len(self.objChildren(obj))

    def removeRows(self,row,count,parent):
        """
        :param row: int offset of first row to remove
        :param count: int quantity of rows to remove
        """
        if self.row_manager is None or self.row_manager.fixed_rows:
            return False
        else:
            self.beginRemoveRows(parent,row,row+count-1)
            for i in range(count):
                row_obj = self.row_object(row, parent.internalPointer())
                if row_obj in self.row_list:
                    self.row_manager.delete_row(row_obj)
            self.endRemoveRows()
            return True

    def columnWidthRatios(self):
        return dict([(i, self.columns[i].width) for i in range(len(self.columns))])
    
    def headerData(self, col, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return attrLabel(self.cls, self.columns[col])
        return None

    def data(self, index, role):
        c = self.columns[index.column()]
        r = self.row_object_by_index(index)
        if not index.isValid():
            return None

        primary_value = getattr(r,c.attr if hasattr(c, "attr") else c)
        primary_type = attrType(self.cls, c)
        primary_yoke = attrYoke(self.cls, c)

        if role == QtCore.Qt.DisplayRole:
            if primary_yoke != 'bool':
                return toQType(primary_value,suggested=primary_type)
        elif role == QtCore.Qt.EditRole:
            return toQType(primary_value,suggested=primary_type)
        elif role == QtCore.Qt.CheckStateRole:
            if primary_yoke == 'bool':
                return QtCore.Qt.Checked if primary_value else QtCore.Qt.Unchecked
        elif role == QtCore.Qt.TextAlignmentRole:
            if hasattr(r,"textAlignmentRole"):
                return r.textAlignmentRole(c)
            if primary_yoke in ('int','float'):
                return QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter
        elif role == QtCore.Qt.FontRole:
            if hasattr(r,"fontRole"):
                return r.fontRole(c)
        elif role == QtCore.Qt.ForegroundRole:
            if hasattr(r,"foregroundRole"):
                return r.foregroundRole(c)
        elif role == QtCore.Qt.BackgroundRole:
            if hasattr(r,"backgroundRole"):
                return r.backgroundRole(c)
        elif role == QtCore.Qt.BackgroundRole:
                return r.foreground_role(c)
        elif role >= QtCore.Qt.UserRole:
            assert QtCore.Qt.UserRole <= role < QtCore.Qt.UserRole+len(self.columns)
            roles = self.roleNames()
            c = self.columns[role - QtCore.Qt.UserRole]
            return toQType(primary_value,suggested=primary_type)

        return None

    def ensureNoFlipper(self, index):
        if not self.row_manager.fixed_rows and index.row() >= len(self.row_list):
            self.row_manager.join_flipper(self.flipper_rows[0])
            del self.flipper_rows[0]
            if len(self.flipper_rows) == 0:
                self.beginInsertRows(QtCore.QModelIndex(),len(self.row_list),len(self.row_list))
                self.flipper_rows.append(self.row_manager.add_flipper_row())
                self.endInsertRows()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        #print value
        if not index.isValid():
            return None
        #elif role != QtCore.Qt.DisplayRole and role != QtCore.Qt.EditRole:
            #return None

        c = self.columns[index.column()]
        r = self.row_object_by_index(index)
        newValue = fromQType(value, attrType(self.cls, c))
        oldValue = getattr(r, c.attr if hasattr(c, "attr") else c)
        #print attrType(self.cls, c), newValue, oldValue
        #  The None/"" dichotomy is a frightening mess -- we need to think carefully about all that.
        if oldValue is None:
            oldValue = toQType(oldValue, attrType(self.cls, c))

        result = True
        if newValue != oldValue:
            self.ensureNoFlipper(index)

            try:
                setattr(r, c.attr if hasattr(c, "attr") else c, newValue)
            except:
                # TODO:  Consider high-lighting the table cell which failed
                # however, we believe this should be in the view, not the model
                # Note:  We expect a dialog listen on the session message event for session validation messages.
                result = False
        return result

    def flags(self,index):
        result = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        c = self.columns[index.column()]
        if not self.readonly and not attrReadonly(self.cls, c):
            if attrYoke(self.cls, c) == 'bool':
                result |= QtCore.Qt.ItemIsUserCheckable
            else:
                result |= QtCore.Qt.ItemIsEditable
        result |= QtCore.Qt.ItemIsDragEnabled
        if len(self.dropCallable) > 0:
            result |= QtCore.Qt.ItemIsDropEnabled
        return result

    def supportedDropActions(self):
        return QtCore.Qt.ActionMask

    def dropMimeData(self, data, action, row, column, parent):
        x = False
        for type in data.formats():
            if self.dropCallable.has_key(type):
                x = self.dropCallable[type](data, action, row, column, parent)
        return x

    def mimeTypes(self):
        return self.dropCallable.keys() + ["text/plain"]

    def mimeData(self, indexes):
        result = modelMimeRectangle(self, indexes)
        if self.objectConverter is not None and hasattr(self.objectConverter, "mime"):
            self.objectConverter.mime(result, self, indexes)
        return result


def modelMimeRectangle(model, selected):
    """
    Fill and return a QMimeData with the text in the indexes listed.  The copied 
    data is a rectangular array of the selected cells.  Cells not selected that 
    fall with-in a row or column of the enclosing rectangle, are indicated only by 
    an empty string in the tab delimited grid.
    """
    def index_row_lineage(index):
        r = []
        while index.isValid():
            r.append(index.row())
            index = index.parent()
        r.reverse()
        return tuple(r)

    def row_data(full_cols, my_indexes):
        result = []
        my_cols = dict([(i.column(), i) for i in my_indexes])
        for x in full_cols:
            r = ""
            if x in my_cols:
                r = str(fromQType(model.data(my_cols[x], QtCore.Qt.DisplayRole)))
            result.append(r)
        return result

    m = QtCore.QMimeData()
    rows = []
    if len(selected):
        sortable_index_list = [(index_row_lineage(i), i.column(), i) for i in selected]
        cols = list(set([x[1] for x in sortable_index_list]))
        cols.sort()

        rows = list(set([x[0] for x in sortable_index_list]))
        rows.sort()

        textrows = ['\t'.join([model.headerData(x,QtCore.Qt.Horizontal,QtCore.Qt.DisplayRole) for x in cols])]
        for y in rows:
            my_indexes = [x[2] for x in sortable_index_list if x[0] == y]
            textrows.append('\t'.join(row_data(cols, my_indexes)))
    m.setData("text/plain",'\n'.join(textrows))
    return m


class WidgetAttributeMapper(object):
    """
    This class provides a hook between widgets in the graphical toolkit and 
    attributes on a python object.

    The widgets created are selected by the sub-classes of InputYoke which 
    are assigned in the yokes_dict function in input_yoke.py.  TODO:  We 
    aim to expose the initialization of this dictionary to the application 
    programmer to allow widget types to be entirely replaced through-out the 
    application.

    It is encouraged that all widgets be created dynamically as Qt's layout 
    engine is extremely robust and configurable.  However, the bind member 
    function of WidgetAttributeMapper provides the ability to take a widget 
    created through other means and connect it to this mapping with a yoke.  
    This may be used with widgets generated via Qt designer tool.
    
    :param cls:  class we are mapping; we look for python descriptors in 
        this class (both UserAttr and sqlalchemy properties)
    :param objectConverter:  a callable taking an instance of type cls and 
        returning an id/handle as appropriate for any bound CommandMenu 
        entities.
    """
    
    def __init__(self, cls, obj=None, objectConverter=None):
        self.cls = cls
        self.obj = obj
        self.objectConverter = objectConverter
        self.widgets = []
        self.yokes = []
        self.attribute_map = {}

    def connect_instance(self,obj):
        self.obj = obj
        if issubclass(self.cls, ModelObject):
            instanceEvent(self.obj, "set", "")(self.row_object_change)
        for y in self.yokes:
            y.Bind()

    def reverse_yoke(self,attr,yoke):
        if attr not in self.attribute_map:
            self.attribute_map[attr] = []
        self.attribute_map[attr].append(yoke)

    def focusAttribute(self, attr):
        """
        Set focus the input yoke corresponding to the attribute attr.
        
        If attr is not a mapped attribute, this method does nothing.
        """
        if attr in self.attribute_map:
            y = self.attribute_map[attr]
            if len(y) > 0:
                y[0].focus()

    def attributeWidget(self, attr):
        """
        Returns the main qt widget in the yoke corresponding to attr.  The 
        exact meaning of the return of this function is subject to the 
        yoke implementation.
        
        If attr is not a mapped attribute, this method returns None.
        """
        if attr in self.attribute_map:
            y = self.attribute_map[attr]
            if len(y) > 0:
                return y[0].mainWidget()
        return None

    def getObjectAttr(self,attr):
        return getattr(self.obj,attr)

    def setObjectAttr(self,attr,value):
        setattr(self.obj,attr,value)

    def row_object_change(self, obj, attr, oldvalue):
        assert obj is self.obj
        if attr not in self.attribute_map:
            return
        for y in self.attribute_map[attr]:
            y.Update(attr)

    def addBoundForm(self,layout,attrs):
        """
        Creates a QFormLayout and adds a collection of labeled fields.
        
        :param attrs:  a list of property names of the class specified with `cls` in __init__.
        """
        grid = LayoutLayout(layout,QtGui.QFormLayout())

        for a in attrs:
            self.addBoundField(grid,a)

    def labelAlignment(self,layout,labelPos):
        parent = layout.parentWidget()
        if parent is None:
            style = QtGui.QApplication.style()
        else:
            style = parent.style()
        if style is not None and labelPos == "l":
            return QtCore.Qt.Alignment(style.styleHint(QtGui.QStyle.SH_FormLayoutLabelAlignment))
        return None

    def addBoundFieldGrid(self,grid,attr,row,column,rowSpan=1,columnSpan=1,labelPos="left"):
        """
        :param labelPos:  label can be one of the following
         - left:  label in cell (row,column) and edit widget in cell (row,column+1)
         - above, top:  label in cell (row,column) and edit widget in cell (row,column+1)
         - right:  edit widget in cell (row,column) and label in cell (row+1,column)
         - below, bottom:  edit widget in cell (row,column) and label in cell (row,column+1)
         - None, '':  no label and edit widget in cell (row,column)
        :param rowSpan:  The span of 
        """
        import input_yoke

        y = input_yoke.YokeFactory(self.cls,attr, self)
        self.yokes.append(y)
        w = y.Factory()

        if labelPos not in ['',None]:
            label_key = labelPos[0].lower()
            label_cell = {'l': (row,column), 'a': (row,column), 't': (row,column), 'r': (row,column+1), 'b': (row+1,column)}.get(label_key)
            widget_cell = {'l': (row,column+1), 'a': (row+1,column), 't': (row+1,column), 'r': (row,column), 'b': (row,column)}.get(label_key)

            label = QtGui.QLabel(attrLabel(self.cls,attr))
            align = self.labelAlignment(grid,label_key)
            if align is not None:
                label_cell += (align,)
            grid.addWidget(label, *label_cell)
            label.setBuddy(w)
        else:
            widget_cell = (row,column)
        widget_cell_span = widget_cell + (rowSpan, columnSpan)
        grid.addWidget(w, *widget_cell_span)
        return w

    def addBoundField(self,layout,attr):
        """
        Adds a labeled field to a layout.
        """
        import input_yoke

        y = input_yoke.YokeFactory(self.cls,attr, self)
        self.yokes.append(y)
        w = y.Factory()

        if isinstance(layout,QtGui.QFormLayout):
            if y.LabelStyle() == input_yoke.InputYoke.Label.Internal:
                layout.addRow("",w)
            else:
                layout.addRow(attrLabel(self.cls,attr),w)
        else:
            layout.addWidget(w)
        
        return w

    def addColumn(self,attr):
        colIndex = -1
        for i in range(len(self.columns)):
            if self.columns[i] == attr:
                colIndex = i
        if colIndex == -1:
            colIndex = len(self.columns)
            self.columns.append(attr)
        return colIndex

    def submit(self):
        for y in self.yokes:
            y.Save()

    def bind(self, widget, attr, yoke_factory=None):
        import input_yoke

        if yoke_factory is None:
            y = input_yoke.YokeFactory(self.cls, attr, self)
        else:
            y = yoke_factory(self, attr)
        y.AdoptWidget(widget)
        self.yokes.append(y)

class MapperMixin(object):
    """
    MapperMixin holds a list of :class:`WidgetAttributeMapper` classes and 
    provides helpers to ensure that the whole collection is synchronized before save.
    """
    def init_mapper_list(self):
        if not hasattr(self,"_mapper_list"):
            self._mapper_list = []

    def mapClass(self, cls, loaderTarget=None, objectConverter=None):
        """
        This function creates a new :class:`WidgetAttributeMapper`, adds it 
        to the list of mappers and returns this new instance.
        
        In addition, if loaderTarget is specified, a callable loader function is
        constructed and passed to self.addLoader.  It is assumed that we are
        mixed in with some class that has loader callbacks.  Such an
        implemenation is :class:`qtalchemy.dialogs.BoundDialog`.
        
        :param loaderTarget:  a callable taking the dataContext of the ambient
            dialog and returning an object instance to connect to the return
            WidgetAttributeMapper
        :param objectConverter:  a callable converting the object bound to the
            WidgetAttributeMapper to an identifier primary key.
        """
        self.init_mapper_list()
        m = WidgetAttributeMapper(cls, objectConverter=objectConverter)
        self._mapper_list.append(m)
        
        if loaderTarget is not None:
            # this presumes we're in a BoundDialog (or something that quacks like one)
            self.addLoader(lambda dataContext: m.connect_instance(loaderTarget(dataContext)))
        
        return m

    def mapper(self,obj):
        self.init_mapper_list()
        m = WidgetAttributeMapper(None,obj)
        self._mapper_list.append(m)
        return m

    def submit(self):
        if not hasattr(self,"_mapper_list"):
            return

        for m in self._mapper_list:
            m.submit()



# -*- coding: utf-8 -*-
"""
We test the basic UserAttr based events.
"""

from qtalchemy import UserAttr, ModelObject

class Line_A_Test(ModelObject):
    classEvents = ModelObject.Events()
    
    quantity = UserAttr(float,"Quantity")
    unit = UserAttr(float,"Unit Price")
    price = UserAttr(float,"Price")

    @classEvents.add("set", "quantity")
    @classEvents.add("set", "unit")
    def on_change_unit(self, attr, oldvalue):
        self.price = self.quantity * self.unit

class Line_B_Test(ModelObject):
    quantity = UserAttr(float,"Quantity")
    unit = UserAttr(float,"Unit Price")
    price = UserAttr(float,"Price")

    @price.on_get
    def on_get_price(self):
        return self.quantity * self.unit

def on_change_test():
    line = Line_A_Test()
    line.quantity = 3
    line.unit = 1.25
    assert abs(line.price - 3.75)<.001, "The price should be quantity*unit"

def on_get_test():
    line = Line_B_Test()
    line.quantity = 3
    line.unit = 1.25
    assert abs(line.price - 3.75)<.001, "The price should be quantity*unit"



# -*- coding: utf-8 -*-

from auth_settings import AuthSettings
from auth import LoginDialog, auth_dialog
from multiauth import MultiAuthDialog, multi_auth_dialog

from objectforms import MapperDialog, BoundDialog, BoundDialogObject, \
                        messaged_commit, exception_message_box, appMessage, SessionMessageBox

try:
    from PyQt4 import QtDeclarative
    from quicklist import QuickList
except ImportError, e:
    pass



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
from qtalchemy import *
from PyQt4 import QtCore, QtGui
from auth_settings import AuthSettings
from objectforms import MapperDialog, exception_message_box

class MultiAuthDialog(MapperDialog):
    def __init__(self,parent=None,init_session_maker=None,init_db=None):
        MapperDialog.__init__(self,parent)
        self.init_session_maker = init_session_maker
        self.init_db = init_db

        self.settings = AuthSettings()
        self.settings.loadSettings()

        self.setWindowTitle(self.tr("Log In"))

        self.m = self.mapClass(AuthSettings)
        
        main = QtGui.QVBoxLayout(self)
        self.tabs = LayoutWidget(main,QtGui.QTabWidget())
        self.tabs.addTab(self.sqlite_tab(),"SQLite")
        self.tabs.addTab(self.firebird_tab(),"Firebird")
        self.tabs.addTab(self.mysql_tab(),"MySQL")
        self.tabs.addTab(self.postgresql_tab(),"PostgreSQL")
        for i in range(4):
            if self.tabs.widget(i).prefix == self.settings.server_type:
                self.tabs.setCurrentIndex(i)
                if self.settings.remember_user and hasattr(self.tabs.widget(i),"password"):
                    self.tabs.widget(i).password.setFocus()
                elif self.settings.remember_server and hasattr(self.tabs.widget(i),"user"):
                    self.tabs.widget(i).user.setFocus()

        remember_layout = LayoutLayout(main,QtGui.QHBoxLayout())
        self.m.addBoundField(remember_layout,"remember_server")
        self.m.addBoundField(remember_layout,"remember_user")
        self.m.connect_instance(self.settings)

        buttonbox = LayoutWidget(main,QtGui.QDialogButtonBox())
        buttonbox.addButton(QtGui.QDialogButtonBox.Ok)
        buttonbox.addButton(QtGui.QDialogButtonBox.Cancel)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)

    def sqlite_tab(self):
        wid = QtGui.QWidget()
        wid.prefix = "sqlite"
        wid.m = self.mapClass(AuthSettings)
        l1 = QtGui.QVBoxLayout(wid)
        selector = LayoutLayout(l1,QtGui.QHBoxLayout())
        LayoutWidget(selector, QtGui.QLabel("Database File:"))
        self.sqlite_file = wid.m.addBoundField(selector,"database")
        LayoutWidget(selector, QtGui.QPushButton("&Browse...")).clicked.connect(self.sqlite_browse)
        
        if self.init_db is not None:
            init = LayoutLayout(l1,QtGui.QHBoxLayout())
            init.addStretch()
            LayoutWidget(init, QtGui.QPushButton("&Initialize Database")).clicked.connect(self.sqlite_initialize)
        
        wid.m.connect_instance(self.settings)

        return wid

    def sqlite_browse(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self,"Select Existing SQLite Database",filter="SQLite databases (*.db);;All Files (*.*)")
        if fileName != "":
            self.sqlite_file.setText(fileName)

    def sqlite_initialize(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self,"New SQLite Database",filter="SQLite databases (*.db);;All Files (*.*)")
        if fileName != "":
            self.init_db("sqlite:///" + fileName)
            self.sqlite_file.setText(fileName)

    def mysql_tab(self):
        wid = QtGui.QWidget()
        wid.prefix = "mysql"
        wid.m = self.mapClass(AuthSettings)
        grid = QtGui.QFormLayout(wid)
        wid.server = wid.m.addBoundField(grid,"server")
        wid.m.addBoundField(grid,"database")
        wid.user = wid.m.addBoundField(grid,"user")
        wid.password = wid.m.addBoundField(grid,"password")
        wid.password.setEchoMode(QtGui.QLineEdit.Password)
        wid.m.connect_instance(self.settings)

        return wid

    def postgresql_tab(self):
        wid = QtGui.QWidget()
        wid.prefix = "postgres"
        wid.m = self.mapClass(AuthSettings)
        grid = QtGui.QFormLayout(wid)
        wid.server = wid.m.addBoundField(grid,"server")
        wid.m.addBoundField(grid,"database")
        wid.user = wid.m.addBoundField(grid,"user")
        wid.password = wid.m.addBoundField(grid,"password")
        wid.password.setEchoMode(QtGui.QLineEdit.Password)
        wid.m.connect_instance(self.settings)

        return wid

    def firebird_tab(self):
        wid = QtGui.QWidget()
        wid.prefix = "firebird"
        wid.m = self.mapClass(AuthSettings)
        grid = QtGui.QFormLayout(wid)
        wid.server = wid.m.addBoundField(grid,"server")
        wid.m.addBoundField(grid,"database")
        wid.user = wid.m.addBoundField(grid,"user")
        wid.password = wid.m.addBoundField(grid,"password")
        wid.password.setEchoMode(QtGui.QLineEdit.Password)
        wid.m.connect_instance(self.settings)

        return wid

    def accept(self):
        self.connection = ""
        self.tabs.currentWidget().m.submit()
        self.m.submit()
        self.settings.saveSettings()

        prefix = self.tabs.currentWidget().prefix
        if prefix == "sqlite":
            self.connection = "sqlite:///%s" % (self.sqlite_file.text(),)
        else:
            self.connection = "%s://%s:%s@%s/%s" % (self.settings.server_type, self.settings.user, self.settings.password, self.settings.server, self.settings.database)
        try:
            #print self.connection
            if self.init_session_maker is not None:
                self.init_session_maker(self.connection)
            super(MultiAuthDialog,self).accept()
        except Exception, e:
            exception_message_box(e,"There was an error initializing the data.",icon=QtGui.QMessageBox.Warning)

def multi_auth_dialog(parent=None,init_session_maker=None,init_db=None):
    a = MultiAuthDialog(parent,init_session_maker,init_db)
    a.show()
    return a.exec_()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
Attach command descriptions and short-cut keys to python functions.  These 
command sets can be converted into QMenu and QToolbar objects.  Additionally, 
one can fill QDialogButtonBox classes.  Facilities are included for classifying 
and sorting the commands for global application consistency.
"""

import weakref
import inspect
from PyQtHelpers import *
from PyQt4 import QtCore, QtGui

DROP_ACTION_TYPE="drop-action"

class Command(object):
    """
    This is a decorator for entity methods to mark them as methods that should be included 
    in the GUI command set in places with this entity.  This enables automatic generation of 
    button boxes and tool-bars.

    See :class:`BoundCommandMenu` for more details.
    """
    def __init__(self, descr, default=False, viewRelated=False, requireSelection=True, needs_reload=True, key=None, statusTip=None, iconFile=None, type="action", sort=500, mimeType=None):
        """
        :param descr: Text description of command for menu items
        :param default: Default command for item activation.
        :param viewRelated: True if this command must be activated in the presence of a view
        :param requireSelection: Set to false if the command function makes sense with-out a current object
        :param needs_reload: Set to false if the command does not modify the database in a way that would require refreshing the screen.
        :param key: shortcut key as given by QtCore.Qt.Key enumeration
        :param type: classifies the command for menu structure - new/view/action/delete are common options with special meanings to the sort algorithm
        """
        self.descr = descr
        self.viewRelated = viewRelated
        self.requireSelection = requireSelection
        self.default = default
        self.needs_reload = needs_reload
        self.iconFile = iconFile
        self.key = key
        self.statusTip = statusTip
        self.command_type = type
        self.sort = sort
        self.func = None

    def __call__(self, func):
        func._command_me = self
        func.descr = self.descr
        self.func = func
        return func

    def name(self):
        return self.func.func_name

    def action(self, parent):
        # drag & drop operations are not prompted by QAction gui events (and they must come with mime data)
        if self.command_type == DROP_ACTION_TYPE:
            return None

        action = QtGui.QAction(self.descr, parent)
        if self.key is not None:
            action.setShortcut(self.key)
        if self.statusTip is not None:
            action.setStatusTip(parent.tr(self.statusTip))
        if self.iconFile is not None:
            action.setIcon(QtGui.QIcon(self.iconFile))
        return action

    def sort_key(self):
        assert 0<=self.sort<1000, "We expect you to sort between 0 and 999 inclusive"
        action_map = {"new": 2, "view": 4, "action": 6, "delete": 8}
        try:
            actInt = action_map[self.command_type]
        except KeyError, e:
            actInt = 5
        return "{0:01f}{1:03f}".format(actInt, self.sort)

class CommandMenu(object):
    """
    The CommandMenu class defines a list of menu commands which can be incorporated 
    into menus or toolbars associated with some UI object.  Multiple CommandMenu 
    objects can be associated with a single class.  This enables modularity in 
    the menu structure.
    
    The principle interaction with CommandMenu is a method decorator 
    :func:`command` to add a command to the menu.  Additional decorators 
    are :func:`itemNew`, :func:`itemAction` and :func:`itemDelete` which 
    are useful with the :func:`BoundCommandMenu.withView`.
    """
    def __init__(self, name):
        self.commands = []
        self.name = name

    def command(self, descr, key=None, statusTip=None, type="action", iconFile=None, sort=500, default=False, requireSelection=False, viewRelated=False, needs_reload=True):
        """
        :param descr:  string that appears in the menu (or other UI element)
        :param key: short-cut key to apply to this command
        :param type:  a string value typically chosen from "new", "action", "delete"; 
            This is used to sort the menu.
        :param sort:  integer from 0 to 1000 used to fine-tune the sort after the 'type'

        A function decorated with itemNew must have one parameter:
        
        #.  command object (probably self)
        """
        def decorator(f):
            x = Command(descr, 
                key=key, 
                type=type, 
                statusTip=statusTip, 
                iconFile=iconFile, 
                sort=sort, 
                default=default, 
                viewRelated=viewRelated, 
                requireSelection=requireSelection, 
                needs_reload=needs_reload)
            x.func = f
            self.commands.append(x)
            self.commands.sort(key=lambda x:x.sort_key())
            return f
        return decorator

    def itemNew(self, **kwargs):
        """
        A function decorator similar to :func:`command` but with parameters 
        defaulted appropriate for adding an item to an adjacent view.
        
        A function decorated with itemNew must have two parameters:
        
        #.  command object (probably self)
        #.  identifier or object acted on.  This should have a default of 
            None if requireSelection is not specified in itemNew.
        """
        if not kwargs.has_key("descr"):
            kwargs["descr"] = "&New..."
        if not kwargs.has_key("iconFile"):
            kwargs["iconFile"] = ":/qtalchemy/default-new.ico"
        if not kwargs.has_key("requireSelection"):
            kwargs["requireSelection"] = False
        if not kwargs.has_key("viewRelated"):
            kwargs["viewRelated"] = True
        kwargs["type"] = "new"
        return self.command(**kwargs)

    def itemDelete(self, **kwargs):
        """
        A function decorator similar to :func:`command` but with parameters 
        defaulted appropriate for deleting an item in an adjacent view.
        
        A function decorated with itemNew must have two parameters:
        
        #.  command object (probably self)
        #.  identifier or object acted on.  This should have a default of 
            None if requireSelection is not specified in itemNew.
        """
        if not kwargs.has_key("descr"):
            kwargs["descr"] = "&Delete..."
        if not kwargs.has_key("requireSelection"):
            kwargs["requireSelection"] = True
        if not kwargs.has_key("viewRelated"):
            kwargs["viewRelated"] = True
        if not kwargs.has_key("iconFile"):
            kwargs["iconFile"] = ":/qtalchemy/default-delete.ico"
        kwargs["type"] = "delete"
        return self.command(**kwargs)

    def itemAction(self, descr, **kwargs):
        """
        A function decorator similar to :func:`command` but with parameters 
        defaulted appropriate for deleting an item in an adjacent view.
        
        A function decorated with itemNew must have two parameters:
        
        #.  command object (probably self)
        #.  identifier or object acted on.  This should have a default of 
            None if requireSelection is not specified in itemNew.
        """
        if not kwargs.has_key("requireSelection"):
            kwargs["requireSelection"] = True
        if not kwargs.has_key("viewRelated"):
            kwargs["viewRelated"] = True
        kwargs["type"] = "action"
        return self.command(descr, **kwargs)

    def dropCommand(self, mimeType, **kwargs):
        """
        A command function for when a bit of mime-data is dropped on 
        to the row.  This is typically used with-in the context of a 
        QAbstractItemView derived widget.
        """
        if not kwargs.has_key("requireSelection"):
            kwargs["requireSelection"] = True
        kwargs["viewRelated"] = True
        kwargs["type"] = DROP_ACTION_TYPE
        return self.command(mimeType, **kwargs)

    def __get__(self, instance, owner):
        try:
            return getattr(instance, self.name)
        except AttributeError as e:
            setattr(instance, self.name, BoundCommandMenu(instance, self))
            return getattr(instance, self.name)

class DomainEntity(CommandMenu):
    """
    DomainEntity is deprecated, but still used by the ForeignKey lookup stuff

    The DomainEntity object defines how an object appears and can be manipulated by the user.  This has 
    several components::
    
    * Defining database queries to provide user view of lists of these domain entities.
    * Key field information to map back to individual tables.
    * Commands acting on a record in the principle table (identified by primary key).
    
    QtAlchemy takes the view that rows of database queries do not map well to typical object oriented 
    design methods.  The DomainEntity provides a more abstract view of commands associated to an entity which 
    can be bound to arbitrary database queries with-in the code.  Thus a joined query can actually offer user 
    commands associated with any of the tables in the query.
    """

    def __init__(self, info={}):
        """
        The following assignments are a common proto-type associating this DomainEntity with a 
        explicit table.  We do not yet define the sql joins clearly.
        """
        CommandMenu.__init__(self,"base_menu")
        self.key_column = None
        self.list_display_columns = None
        self.list_search_columns = None
        self.info = info

    def list_query_converter(self):
        from sqlalchemy.orm import Query
        queryCols = tuple([self.key_column.label("_hidden_id")] + self.list_display_columns)
        return (Query(queryCols).order_by(self.list_display_columns[0], self.key_column), lambda x: x._hidden_id)


class CommandEvent(object):
    """
    This class is passed with the `preCommand` signal, optionally to the command 
    function, and to the `refresh` signal after the command function is executed.  
    The command functions, decorated by members of :class:`CommandMenu`, are 
    expected to indicate refresh context in a semi-standard way which isn't 
    entirely known yet.
    
    Recipients of the `preCommand` signal can :func:`abort` the command if they 
    wish; this is useful if the ambient screen could not save its data for 
    some reason.  The CommandEvent class passed could be pre-loaded with settings 
    which might be useful with-in the command function.
    
    Recipients of the `refresh` signal may possibly desire to reload from 
    context clues put into the passed CommandEvent instance by the command 
    function.
    
    Note that the free-form nature of screen context can create a tight coupling 
    between the screens using the associate :class:`CommandMenu` and the the 
    command functions themselves.
    
    This class works can work in tight conjunction with :class:`qtalchemy.dialogs.BoundDialog`.  
    In this context, the preCommandSave function will set a member savedDataContext
    with the current main_row object bound to the BoundDialog.
    """
    def __init__(self, cmd=None):
        self.cmd = cmd
        self.aborted = False
    
    def abort(self):
        """
        Call this method of the event parameter to the preCommand signal to 
        stop the command function from being fired.  If aborted, the refresh 
        signal will not be emitted.
        """
        self.aborted = True

class BoundCommandMenuItem(object):
    """
    The class :class:`BoundCommandMenu` is responsible for creating and managing instances of this class.
    
    This class binds comamnds from an entity to a PyQt/PySide view where each row represents an 
    entity to the entity class.  The proxy slots 'activatedProxy' and 'triggeredProxy' redirect 
    to the specified function of the entity instance.
    """
    def __init__(self, cmd, action, objectConverter, binding):
        """
        :param cmd:  a :class:`Command` object
        :param action:  A QAction or QPushButton
        :param objectConverter:  A callable taking row objects (QModelIndex.internalPointer()) to id's for the entity functions.
            If objectConverter is None, then use the default model objectConverter.
        """
        self.cmd = cmd
        self.action = action
        self.objectConverter = objectConverter
        self.binding = binding

        if binding.view is not None:
            self.model = binding.view.model()
            self.item_sel = binding.view.selectionModel()

            # hook to QItemSelectionModel::selectionChanged signal for enabling/disabling
            # hook to QAbstractItemModel::modelReset signal for enabling/disabling
            self.item_sel.selectionChanged.connect(self.enable_buttons)
            self.model.modelReset.connect(self.enable_buttons)

            if self.cmd.command_type == DROP_ACTION_TYPE:
                self.model.dropCallable[self.cmd.descr] = self.dropAction

        if isinstance(self.action, QtGui.QPushButton):
            self.action.clicked.connect(self.triggeredProxy)
        elif isinstance(self.action, QtGui.QAction):
            self.action.triggered.connect(self.triggeredProxy)

    def enable_buttons(self, sel=None, desel=None):
        # setEnabled is sufficient for both QPushButton and QAction
        if self.cmd.requireSelection and self.action is not None:
            indexes = self.binding.view.selectedIndexes()
            self.action.setEnabled(len(indexes)>0)

    def dropAction(self, data, action, row, column, parent):
        event = CommandEvent(self.cmd)
        self.binding.preCommand.emit(event)
        if event.aborted:
            return
        
        # this is a bit of a hack to preCommand first:
        # it is presumed that preCommand saves the data so that the database keys are up-to-date
        handle = parent.internalPointer()
        if handle is not None:
            handle = self.model.objectConverter(handle)
        args = inspect.getargspec(self.cmd.func).args
        d = {}
        if 'event' in args:
            d['event'] = event
        self.cmd.func(self.binding.target, handle, str(data.data(self.cmd.descr)), **d)
        self.binding.refresh.emit(event)
        if self.cmd.needs_reload:
            if self.binding.view:
                self.model.reset_content_from_session()
        return not event.aborted

    def objectActivatedProxy(self, obj=None, objectConverter=None):
        # self.objectConverter takes precedence!
        objectConverter = objectConverter if self.objectConverter is None else self.objectConverter
        event = CommandEvent(self.cmd)
        self.binding.preCommand.emit(event)
        if event.aborted:
            return
        
        # this is a bit of a hack to preCommand first:
        # it is presumed that preCommand saves the data so that the database keys are up-to-date
        handle = obj
        if handle is not None:
            handle = objectConverter(handle)
        args = inspect.getargspec(self.cmd.func).args
        d = {}
        if 'event' in args:
            d['event'] = event
        self.cmd.func(self.binding.target, handle, **d)
        self.binding.refresh.emit(event)
        if self.cmd.needs_reload:
            if self.binding.view:
                self.model.reset_content_from_session()

    def activatedProxy(self, index):
        if self.cmd.requireSelection and index is None:
            import warnings
            warnings.warn("No selected view index for {0}".format(self.cmd.name))
        else:
            if self.cmd.viewRelated:
                handle = index
                if handle is not None:
                    handle = handle.internalPointer()
                self.objectActivatedProxy(handle, self.model.objectConverter)
            else:
                event = CommandEvent(self.cmd)
                self.binding.preCommand.emit(event)
                if event.aborted:
                    return
                try:
                    args = inspect.getargspec(self.cmd.func).args
                except Exception as e:
                    args = ()
                d = {}
                if 'event' in args:
                    d['event'] = event
                self.cmd.func(self.binding.target, **d)
                self.binding.refresh.emit(event)

    def triggeredProxy(self, checked=False):
        if self.binding.obj is not None:
            obj = self.binding.obj
            self.objectActivatedProxy(obj)
        elif self.binding.mapper is not None:
            self.objectActivatedProxy(self.binding.mapper.obj, self.binding.mapper.objectConverter)
        elif self.binding.view is not None:
            indexes = self.binding.view.selectedIndexes()
            index = None if len(indexes)==0 else indexes[0]
            self.activatedProxy(index)
        else:
            self.activatedProxy(None)

class BoundCommandMenu(QtCore.QObject):
    """
    This class holds the :class:`BoundCommandMenuItem` objects binding an entity to a view.  This is 
    the main class for tying :class:`CommandMenu` instances to PyQt/PySide views.
    
    Primarily the work of this class is carried out in the :class:`BoundCommandMenuItem` objects which holds 
    slots connected to signals from the view.  This class also attaches the actions to the view.
    
    Note that if the view is a :class:`qtalchemy.widgets.TableView`, the 
    default context menu policy of :class:`qtalchemy.widgets.TableView` is 
    QtCore.Qt.ActionsContextMenu.
    """
    preCommand = Signal(object, name='preCommand')
    refresh = Signal(object, name='refresh')

    def __init__(self, target, menu, view=None, mapper=None, obj=None, webView=None, javaScriptName=None, buttonBox=None, bindDefault=False, objectConverter=None):
        QtCore.QObject.__init__(self)
        self.target = target
        self.menu = menu
        self.view = view
        self.mapper = mapper
        self.obj = obj
        self.webView = webView
        self.javaScriptName = javaScriptName
        self.buttonBox = buttonBox
        self.bindDefault = bindDefault
        self.objectConverter = objectConverter
        
        if self.webView:
            # Is loadFinished soon enough?  What if the user is clicking 
            # links and icons before the load is finished?
            # Note that loadStarted was too soon and the frame was overwritten.
            self.webView.loadFinished.connect(lambda x: self.addJavascriptObjects())
        
        if self.view:
            # we need to hook up to the context menu immediately
            self.resetActions()

    def withView(self, view, buttonBox=None, bindDefault=False, objectConverter=None):
        """
        Return a duplicate of this BoundMenuCommand augmented with a view.  The 
        QAction objects will enable and disable as appropriate for the connected view.
        
        :param view: a QAbstractItemView attached to a QAbstractItemModel
        :param bindDefault: bindDefault can take several values - True, "double", "activated", False (True maps to "activated")
        """
        return BoundCommandMenu(self.target, self.menu, view=view, buttonBox=buttonBox, bindDefault=bindDefault, objectConverter=objectConverter)

    def withMapper(self, mapper, buttonBox=None, bindDefault=False, objectConverter=None):
        """
        Return a duplicate of this BoundMenuCommand augmented with a view.  The 
        QAction objects will enable and disable as appropriate for the connected view.
        
        :param view: a QAbstractItemView attached to a QAbstractItemModel
        """
        return BoundCommandMenu(self.target, self.menu, mapper=mapper, buttonBox=buttonBox, bindDefault=bindDefault, objectConverter=objectConverter)

    def withWebView(self, webView, javaScriptName=None, objectConverter=None):
        """
        Return a duplicate of this BoundMenuCommand augmented with a QWebView.  
        The methods :func:`htmlJavascript`, :func:`htmlIcons` and :func:`htmlButtons` can 
        be used to construct html which calls back to this view.
        
        The loadFinished signal of the passed QWebView is connected and the 
        QWebFrame method addToJavaScriptWindowObject is called to associated 
        this object with name in the javascript name space.
        
        :param webView:  the webview which is associated with the returned object
        :param javaScriptName:  a name for the returned object in the 
            javascript namespace.  If None, a unique name is auto-generated.
        """
        if javaScriptName is None:
            import uuid
            javaScriptName = "bound_command_{0}".format(uuid.uuid1().hex)
        return BoundCommandMenu(self.target, self.menu, webView=webView, javaScriptName=javaScriptName, objectConverter=objectConverter)

    def withObject(self, obj, buttonBox=None, objectConverter=None):
        """
        Return a duplicate of this BoundMenuCommand augmented with a view.  The 
        QAction objects will enable and disable as appropriate for the connected view.
        
        :param view: a QAbstractItemView attached to a QAbstractItemModel
        """
        return BoundCommandMenu(self.target, self.menu, obj=obj, buttonBox=buttonBox, objectConverter=objectConverter)

    def resetActions(self):
        """
        Regenerate QAction objects from self.target
        """
        self.action_bindings = [] # this is annoying -- I need to hold a reference explicitly
        self.button_bindings = []
        self.default_binding = None

        actionParent = self.view
        if actionParent is None and isinstance(self.target, QtGui.QWidget):
            actionParent = self.target
        if actionParent is None:
            actionParent = self.target.parent

        for cmd in self.menu.commands:
            action = cmd.action(actionParent)
            current = BoundCommandMenuItem(cmd, action, objectConverter = self.objectConverter, binding = weakref.proxy(self))
            self.action_bindings.append(current)
            if action is None:
                continue
            if cmd.default:
                self.default_binding = current

            if self.buttonBox is not None:
                #TODO:  map Command.type to QtGui.QDialogButtonBox.*Role items
                button = self.buttonBox.addButton(cmd.descr, QtGui.QDialogButtonBox.ActionRole)
                self.button_bindings.append(BoundCommandMenuItem(cmd, button, objectConverter = self.objectConverter, binding = weakref.proxy(self)))

        if self.default_binding:
            if self.bindDefault is True or self.bindDefault == "activate":
                self.view.activated.connect(self.default_binding.activatedProxy)
            elif self.bindDefault == "double":
                self.view.doubleClicked.connect(self.default_binding.activatedProxy)

        if self.view is not None:
            for cb in self.action_bindings:
                if cb.cmd.viewRelated and cb.action is not None:
                    self.view.addAction(cb.action)

    def fillMenu(self, menu):
        """
        Take the actions in the list and attach them to a menu.
        """
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        for command in self.action_bindings:
            menu.addAction(command.action)

    def fillToolbar(self, toolbar):
        """
        Take the actions in the list and attach them to a toolbar.
        """
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        for command in self.action_bindings:
            toolbar.addAction(command.action)

    def htmlButtons(self, obj):
        """
        Return a string with html button definitions which call the trigger slot.
        
        :param javaScriptName:  the name given when passing this binding object to addToJavaScriptWindowObject.
        """
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        fstring = """<input type="button" id="evalbutton" value="{descr}" onclick="{callback}" />"""
        btns = [fstring.format(descr=binding.cmd.descr, callback=self.htmlJavascript(binding.cmd.name(), obj)) for binding in self.action_bindings]
        return ' '.join(btns)

    def htmlIcons(self, obj):
        """
        Return a string with html image definitions which call the trigger slot.
        
        :param javaScriptName:  the name given when passing this binding object to addToJavaScriptWindowObject.
        """
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        fstring = """<img src="{icon}" alt="{descr}" onclick="{callback}" />"""
        btns = [fstring.format(descr=binding.cmd.descr, icon=binding.cmd.iconFile, callback=self.htmlJavascript(binding.cmd.name(), obj)) for binding in self.action_bindings]
        return ' '.join(btns)

    def htmlJavascript(self, cmd, obj):
        return "{jsn}.objectActivated('{name}', '{id}')".format(jsn=self.javaScriptName, name=cmd, id=obj)

    def addJavascriptObjects(self):
        self.webView.page().currentFrame().addToJavaScriptWindowObject(self.javaScriptName, self)

    @Slot(str, str)
    def objectActivated(self, cmd, obj):
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        for binding in self.action_bindings:
            if binding.cmd.name() == cmd:
                binding.objectActivatedProxy(obj, self.objectConverter)

    @Slot(str)
    def trigger(self, cmd):
        if not hasattr(self, "action_bindings"):
            self.resetActions()
        for binding in self.action_bindings:
            if binding.cmd.name() == cmd:
                binding.triggeredProxy()

    def __len__(self):
        return len(self.menu.commands)

    def __getitem__(self, index):
        return self.menu.commands[index]



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from PyQt4 import QtCore, QtGui
from qtalchemy import fromQType, modelMimeRectangle, writeTableColumnGeo, readTableColumnGeo
from viewbase import ViewBase

class TreeView(QtGui.QTreeView, ViewBase):
    """
    The TreeView provides a QTreeView override with support for the following elements:

    - row deletion with ctrl+delete for editable models
    - tab skipping read-only columns for editable models
    - clipboard copy of the selected blocks
    - load & save column geometry from user settings

    :param parent:  Qt parent widget for this table
    :param extensionId:  settings location for saving column geometry
    
    Note:  You are encouraged to use the extensionId parameter 
    in setModel rather than this one.  The setModel function determines 
    the available columns so that is a better point to specify the 
    related concern of saving column geometry.
    """
    def __init__(self, parent=None, extensionId=None):
        QtGui.QTreeView.__init__(self,parent)
        self.use_edit_tab_semantics = False
        self.setProperty("ExtensionId", extensionId)
        self.preservedRowState = None

        self.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        self.setEditTriggers(QtGui.QAbstractItemView.AllEditTriggers)

        # Ctrl+Delete should delete the row.
        self.del_key = QtGui.QShortcut(self)
        self.del_key.setKey(QtCore.Qt.CTRL + QtCore.Qt.Key_Delete)
        self.del_key.setContext(QtCore.Qt.WidgetWithChildrenShortcut)
        self.del_key.activated.connect(self.delKeyPressed)

        # hook up header events to save section widths
        header = self.header()
        header.sectionResized.connect(lambda x,y,z: self.saveSections())
        header.sectionAutoResize.connect(lambda x,y: self.saveSections())

    def setModel(self, model, extensionId=None):
        """
        :param model:  Qt model to connect to this table
        :param extensionId:  settings location for saving column geometry
        """
        if extensionId is not None:
            # clobber the parameter to __init__ if specified non-None here
            self.setProperty("ExtensionId", extensionId)

        QtGui.QTreeView.setModel(self, model)
        # hook up model resets to preserve selection across reloads
        self.model().modelAboutToBeReset.connect(self.preReset)
        self.model().modelReset.connect(self.postReset)
        if self.property("ExtensionId") is not None:
            readTableColumnGeo(self, self.property("ExtensionId"))

    def keyPressEvent(self, event):
        """
        Copy key copies selection 
        """
        if event.matches(QtGui.QKeySequence.Copy):
            cb = QtGui.QApplication.instance().clipboard()
            cb.setMimeData(modelMimeRectangle(self.model(), self.selectedIndexes()))
        else:
            QtGui.QTreeView.keyPressEvent(self,event)

    def edit(self, index, trigger, event):
        # At this point this method is overridden strictly as a place to set the flag to initiate the 
        # semantics of moveCursor to skip read-only cells in tabbing.
        inherited = QtGui.QTreeView.edit(self, index, trigger, event)
        if inherited:
            self.use_edit_tab_semantics = True
        return inherited

    def moveCursor(self, cursorAction, modifiers):
        if self.use_edit_tab_semantics == True:
            # Skip read-only cells.
            # We max out at iterating through the equivalent of 3 full rows 
            # because that probably indicates an unexpected situation.
            cell_max = self.model().columnCount(None) * 3
            iterations = 0
            if cursorAction == QtGui.QAbstractItemView.MoveNext:
                index = self.currentIndex()
                model = self.model()
                while True and iterations < cell_max:
                    iterations += 1
                    candidate = self.nextIndex(index)
                    if (model.flags(candidate) & QtCore.Qt.ItemIsEditable) == QtCore.Qt.ItemIsEditable:
                        return candidate
                    index = candidate
            elif cursorAction == QtGui.QAbstractItemView.MovePrevious:
                index = self.currentIndex()
                model = self.model()
                while True and iterations < cell_max:
                    iterations += 1
                    candidate = self.prevIndex(index)
                    if (model.flags(candidate) & QtCore.Qt.ItemIsEditable) == QtCore.Qt.ItemIsEditable:
                        return candidate
                    index = candidate
        return QtGui.QTreeView.moveCursor(self,cursorAction,modifiers)

    def columnIndices(self, index):
        m = self.model()
        return [m.index(i, 0, index) for i in range(m.rowCount(index))]

    def saveRowState(self, index):
        for cIndex in self.columnIndices(index):
            if self.isExpanded(cIndex):
                self.preservedRowState.expanded.append(self.model().objectConverter(cIndex.internalPointer()))
                self.saveRowState(cIndex)

    def preReset(self):
        self.preservedRowState = RowState()
        index = self.selectionModel().currentIndex()
        if index is not None and index.isValid():
            oc = self.model().objectConverter
            self.preservedRowState.current = oc(index.internalPointer())
        self.saveRowState(QtCore.QModelIndex())

    def loadRowState(self, index):
        oc = self.model().objectConverter
        for cIndex in self.columnIndices(index):
            id_ = oc(cIndex.internalPointer())
            if self.preservedRowState.isExpanded(id_):
                self.expand(cIndex)
                self.loadRowState(cIndex)
            if id_ == self.preservedRowState.current:
                self.setCurrentIndex(cIndex)

    def postReset(self):
        if self.preservedRowState is not None:
            # the current index is reset with-in the recursive loadRowState
            # there is a possibility that the parent of the selected index has been reset in some way and the current index wouldn't be found.
            self.loadRowState(QtCore.QModelIndex())


class RowState:
    def __init__(self):
        self.current = None
        self.expanded = []

    def isExpanded(self, id):
        return id in self.expanded



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
This module contains convenience functions used elsewhere in the qtalchemy library.

qtalchemy using API2 of PyQt so we need to enable that before importing PyQt4.  We 
do this here as an example and to prepare for the doc-tests.  Note that it is not
illegal to call sip.setapi twice, but the second call must agree in api version with
the first.

    >>> if qt_bindings == 'PyQt4':
    ...     import sip
    ...     sip.setapi('QString', 2)
    ...     sip.setapi('QVariant', 2)
"""

qt_bindings = 'PyQt4'
if qt_bindings == 'PyQt4':
    # we wish to only touch sip if we're PyQt4 based (otherwise sip is unnecessary)
    import sip
    sip.setapi('QString', 2)
    sip.setapi('QVariant', 2)
from PyQt4 import QtGui,QtCore
import datetime
import decimal
import re

# with this conditional, we abstract away this difference between pyqt and pyside
if QtCore.__name__.startswith("PySide"):
    Slot = QtCore.Slot
    Signal = QtCore.Signal
    Property = QtCore.Property
else:
    Slot = QtCore.pyqtSlot
    Signal = QtCore.pyqtSignal
    Property = QtCore.pyqtProperty

def fromQType(v, suggested=None):
    """
    This function takes a PyQt/PySide type and returns the most faithful representative of the item 
    in a native python type.
    
    (Basic) Examples:
    >>> type(fromQType(toQType(45)))
    <type 'int'>
    >>> fromQType(toQType(datetime.date.today()))==datetime.date.today()
    True
    """
    if hasattr(v, "toPyObject"):  # PyQt QVariant
        if v.isNull():
            return None
        v = v.toPyObject()
    elif hasattr(v, "toPyDateTime"):  # PyQt QDateTime
        if v.isNull():
            return None
        v = v.toPyDateTime()
    elif hasattr(v, "toPyDate"):  # PyQt QDate
        if v.isNull():
            return None
        v = v.toPyDate()
    elif hasattr(v, "toPyTime"):  # PyQt QTime
        if v.isNull():
            return None
        v = v.toPyTime()
    elif hasattr(v, "toPython"):  # PySide object
        if v.isNull():
            return None
        v = v.toPython()

    if suggested in (int, float, decimal.Decimal) and v=="":
        if suggested is decimal.Decimal:
            v = decimal.Decimal('0')
        else:
            v = suggested(0)
    elif suggested is datetime.date and isinstance(v,datetime.datetime):
        # TODO:  dangerous down-cast
        v = datetime.date(v.year,v.month,v.day)
    elif suggested is not None:
        try:
            is_ = isinstance(v,suggested)
        except TypeError as e:
            is_ = False

        if not is_:
            v = suggested(v)

    return v

def toQType(v,suggested=None):
    """
    This function takes a native python object and returns it in the most faithful way as a PyQt/PySide type.  
    For API 2, this means that basic types like str and int are returned unchanged.
    
    One notable caveat is that Qt does not have a decimal type so objects of type decimal.Decimal are returned 
    as strings.  The rationale for this is that maintaining decimal exactness of decimals is almost always the 
    right thing.
    
    Examples:
    >>> type(toQType(12))
    <type 'int'>
    >>> type(toQType("A string"))
    <type 'str'>
    >>> toQType(decimal.Decimal("12.34"))
    '12.34'
    >>> type(toQType(datetime.date(1979,1,9))) is QtCore.QDate
    True

    When the value `v` is None, we return a blank version of the suggested type.

    >>> toQType(None,suggested=str)
    ''
    >>> toQType(None,suggested=int)
    0
    """
    if isinstance(v, datetime.date):
        v = QtCore.QDate(v.year,v.month,v.day)
    elif isinstance(v, decimal.Decimal):
        v = str(v)

    if v is None and suggested is not None:
        if suggested == datetime.date:
            v = QtCore.QDate()
        else:
            # I'm baffled, suggested appears to be passed as a python type
            # but the whole point here is to return Qt type!
            v = suggested()

    return v

def ButtonBoxButton(bb,b,role=None):
    """
    :func:`.ButtonBoxButton` is a convenience function which enables button
    construction and assignment on one line of application code.
    """
    #TODO:  The return from this function can be a button or Role constant ... positively bizzarre
    if role is None:
        bb.addButton(b)
    else:
        bb.addButton(b,role)
    return b

def FormRow(form,label,widget):
    """
    :func:`.FormRow` is a convenience function which enables widget
    construction and assignment on one line of application code.
    """
    form.addRow(label,widget)
    return widget

def LayoutLayout(host,inner):
    """
    :func:`.LayoutLayout` is a convenience function which enable layout
    construction and assignment on one line of application code.
    
    See also :func:`.LayoutWidget` :func:`.FormRow` :func:`.ButtonBoxButton`

        >>> app = qtapp()
        >>> class Item(QtGui.QDialog):
        ...     def __init__(self,parent=None):
        ...         QtGui.QDialog.__init__(self,parent)
        ...         vbox = QtGui.QVBoxLayout(self)
        ... 
        ...         form = LayoutLayout(vbox,QtGui.QFormLayout())
        ...         self.book_edit = FormRow(form,"&Book Title",QtGui.QLineEdit())
        ...         self.author_edit = FormRow(form,"&Author",QtGui.QLineEdit())
        ... 
        ...         self.buttons = LayoutWidget(vbox,QtGui.QDialogButtonBox())
        ...         self.ok = ButtonBoxButton(self.buttons,QtGui.QDialogButtonBox.Save)
        ...         self.cancel = ButtonBoxButton(self.buttons,QtGui.QDialogButtonBox.Cancel)
        >>> d = Item()
        >>> d.exec_()  #doctest: +SKIP
        0
    """
    host.addLayout(inner)
    return inner

def LayoutWidget(layout,widget):
    """
    :func:`.LayoutWidget` is a convenience function which enable widget
    construction and assignment on one line of application code.
    """
    layout.addWidget(widget)
    return widget

def writeTableColumnGeo(table,name):
    settings = QtCore.QSettings()
    settings.beginGroup(name)
    for i in range(table.model().columnCount(None)):
        settings.setValue("Column%02i"%i,table.columnWidth(i))
    settings.endGroup()

def readTableColumnGeo(table,name):
    settings = QtCore.QSettings()
    settings.beginGroup(name)
    ei = table.property("ExtensionId")
    table.setProperty("ExtensionId", None)
    total = sum([table.columnWidth(i) for i in range(table.model().columnCount(None))])
    defaults = None
    defaultTotal = None
    if hasattr(table.model(), "columnWidthRatios"):
        defaults = table.model().columnWidthRatios()
        defaults = [defaults[i] for i in range(table.model().columnCount(None))]
        if None not in defaults:
            defaultTotal = sum(defaults)
    for i in range(table.model().columnCount(None)):
        if settings.value("Column%02i"%i,-1) != -1:
            table.setColumnWidth(i,int(settings.value("Column%02i"%i,-1)))
        else:
            if defaults is not None and defaults[i] is not None:
                table.setColumnWidth(i, int(float(defaults[i]) * total / defaultTotal))
    table.setProperty("ExtensionId", ei)
    settings.endGroup()

def suffixExtId(parent, ext):
    id = parent.property("ExtensionId")
    if id is None:
        id = parent.objectName()
    if id is None:
        return ext
    else:
        return id+"/"+ext

class TabMenuHolder:
    def __init__(self, tab):
        self.tab = tab

        self.actSaveLast = QtGui.QAction("&Reopen to last viewed tab", self.tab)
        self.actSaveLast.setCheckable(True)
        self.actSaveThis = QtGui.QAction("&Reopen to current tab", self.tab)
        self.actSaveThis.setCheckable(True)

        self.tab.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tab.customContextMenuRequested.connect(self.contextMenu)

    def contextMenu(self, pnt):
        tb = self.tab.tabBar()
        if tb.tabAt(pnt) == self.tab.currentIndex():
            self.menu = QtGui.QMenu()
            self.menu.addAction(self.actSaveLast)
            self.menu.addAction(self.actSaveThis)
            self.menu.popup(self.tab.mapToGlobal(pnt))

    def setChecks(self, status):
        self.actSaveLast.setChecked(status == WindowGeometry.TAB_REOPEN_LAST)
        self.actSaveThis.setChecked(status == WindowGeometry.TAB_REOPEN_SPECIFIED)

class WindowGeometry(QtCore.QObject):
    """
    This class saves and restores the size and other geometry artifacts about 
    the passed QWidget.  It hooks the closeEvent by attaching itself as an 
    eventFilter to the passed QWidget.

    Table header geometry should be saved by passing an extensionId to 
    :func:`TableView.setModel` at this point.  This may change in the future.

    The geometry is persisted with QSettings under a name that is determined 
    by one of the following (with first items taking precedence).  This name 
    is determined in __init__ and saved for writing the settings later under 
    the same name.
    
    * name parameter
    * widget.property("ExtensionId")
    * widget.objectName()  (recommended)

    The QTabWidgets in the `tabs` list will have a context menu added on them 
    with options about how the tab selection should be recalled on reload.  They 
    can reopen the last opened, always reopen to a specific tab, or let it as 
    default (probably meaning the first tab is shown).

    :param widget: the QWidget for which to save & restore state
    :param name: optional identifier to associate this in the persistent state
    :param size: save & restore window size (default True)
    :param position: save & restore window position (default True)
    :param splitters: list of splitters to save position
    :param tabs: list of QTabWidgets whose tab selection should be managed/remembered
    """
    TAB_REOPEN_DEFAULT = 0
    TAB_REOPEN_LAST = 1
    TAB_REOPEN_SPECIFIED = 2

    def __init__(self, widget, name=None, size=True, position=True, splitters=None, tabs=None):
        QtCore.QObject.__init__(self, widget)

        self.widget = widget
        self.size = size
        self.position = position
        self.name = name
        if self.name is None:
            self.name = widget.property("ExtensionId")
        if self.name is None:
            self.name = widget.objectName()
        self.splitters = splitters if splitters else []
        self.tabs = tabs if tabs else []
        self.tabs = [TabMenuHolder(t) for t in self.tabs]

        for s_index in range(len(self.splitters)):
            self.splitters[s_index].splitterMoved.connect(lambda pos,index,x=s_index: self.updateSplitter(x,pos,index))

        self.tabActions = []
        for t_index in range(len(self.tabs)):
            self.tabs[t_index].tab.currentChanged.connect(lambda newSel,x = t_index: self.updateTab(x,newSel))
            self.tabs[t_index].actSaveLast.triggered.connect(lambda checked, kind=self.TAB_REOPEN_LAST, x=t_index: self.updateTabLastSave(x, checked, kind))
            self.tabs[t_index].actSaveThis.triggered.connect(lambda checked, kind=self.TAB_REOPEN_SPECIFIED, x=t_index: self.updateTabLastSave(x, checked, kind))

        self.restoreState()

        if hasattr(self.widget, "finished"):
            self.widget.finished.connect(self.finished)
        else:
            self.widget.installEventFilter(self)

    def finished(self, i):
        self.saveState(splitters=False, tabs=False)

    def eventFilter(self, obj, event):
        if obj is self.widget and event.type() == QtCore.QEvent.Close:
            self.saveState(splitters=False, tabs=False)
        return QtCore.QObject.eventFilter(self, obj, event)

    def saveState(self, splitters=True, tabs=True):
        """
        This saves the state of all controlled elements.  Some elements are 
        saved immediately when modified (such as splitters).  Thus we suppress 
        the state saving on close for these elements.

        :param splitters:  pass False to suppress the saving of the splitter 
            state for splitters passed in __init__
        :param tabs:  pass False to suppress the saving of the tab 
            state for tabs passed in __init__
        """
        settings = QtCore.QSettings()
        settings.beginGroup(self.name)
        if self.size and self.position:
            settings.setValue("geometry", self.widget.saveGeometry())
        elif self.size:
            settings.setValue("size", self.widget.size())
        elif self.position:
            settings.setValue("pos", self.widget.pos())
        
        if hasattr(self.widget, "saveState"):
            # I'm probably a QMainWindow
            settings.setValue("windowState", self.widget.saveState())

        if splitters:
            for splitter_index in range(len(self.splitters)):
                settings.setValue(self.splitter_persist_location(splitter_index),self.splitters[splitter_index].saveState())

        if tabs:
            for tab_index in range(len(self.tabs)):
                settings.setValue(self.tab_persist_location(tab_index, "to-reopen"), self.tabs[tab_index].tab.currentIndex())

        settings.endGroup()

    def restoreState(self):
        settings = QtCore.QSettings()
        settings.beginGroup(self.name)
        if self.size and self.position:
            if settings.value("geometry") is not None:
                self.widget.restoreGeometry(settings.value("geometry"))
        elif self.size:
            if settings.value("size") is not None:
                self.widget.resize(settings.value("size"))
        elif self.position:
            if settings.value("pos") is not None:
                self.widget.move(settings.value("pos"))
        
        if hasattr(self.widget, "saveState"):
            # I'm probably a QMainWindow
            if settings.value("windowState") is not None:
                self.widget.restoreState(settings.value("windowState"))
        
        for splitter_index in range(len(self.splitters)):
            state = settings.value(self.splitter_persist_location(splitter_index))
            if state is not None:
                self.splitters[splitter_index].restoreState(state)

        for tab_index in range(len(self.tabs)):
            reopen = int(settings.value(self.tab_persist_location(tab_index, "reopen-status"), self.TAB_REOPEN_LAST))
            self.tabs[tab_index].setChecks(reopen)
            state = settings.value(self.tab_persist_location(tab_index, "to-reopen"))
            if state is not None and reopen != self.TAB_REOPEN_DEFAULT:
                self.tabs[tab_index].tab.setCurrentIndex(int(state))
        settings.endGroup()

    def splitter_persist_location(self, splitter_index):
        splitter_name = self.splitters[splitter_index].objectName()
        if splitter_name in [None, ""]:
            splitter_name = str(splitter_index)
        return "splitter/{0}".format(splitter_name)

    def updateSplitter(self, splitter_index, pos, index):
        settings = QtCore.QSettings()
        settings.beginGroup(self.name)
        settings.setValue(self.splitter_persist_location(splitter_index),self.splitters[splitter_index].saveState())
        settings.endGroup()

    def tab_persist_location(self, tab_index, v):
        tab_name = self.tabs[tab_index].tab.objectName()
        if tab_name in [None, ""]:
            tab_name = str(tab_index)
        return "tab/{0}/{1}".format(tab_name, v)

    def updateTabLastSave(self, tab_index, checked, how):
        reopen = None
        tab_to_reopen = None
        if checked:
            reopen = how
            tab_to_reopen = self.tabs[tab_index].tab.currentIndex()
        else:
            reopen = self.TAB_REOPEN_DEFAULT

        if reopen is not None:
            self.tabs[tab_index].setChecks(reopen)

            settings = QtCore.QSettings()
            settings.beginGroup(self.name)
            settings.setValue(self.tab_persist_location(tab_index, "reopen-status"), reopen)
            if tab_to_reopen is not None:
                settings.setValue(self.tab_persist_location(tab_index, "to-repen"), tab_to_reopen)
            settings.endGroup()

    def updateTab(self, tab_index, newSel):
        settings = QtCore.QSettings()
        settings.beginGroup(self.name)
        if int(settings.value(self.tab_persist_location(tab_index, "reopen-status"), self.TAB_REOPEN_LAST)) == self.TAB_REOPEN_LAST:
            settings.setValue(self.tab_persist_location(tab_index, "to-reopen"), self.tabs[tab_index].tab.currentIndex())
        settings.endGroup()


class OnlineHelp(QtCore.QObject):
    """
    Install help buttons and an F1 event filter both of which direct to a web 
    url with appropriate anchor points (for context specific help).
    
    For context sensitive help, the F1 key is linked to the same URL with an 
    anchor suffix.  The suffix comes from the active control BoundFieldClass 
    property.  This property is set for controls bound by an :class:`InputYoke`.
    
    :param parent: the parent window to which help applies
    :param url: the URL to direct the user to when help is needed
    :param buttonBox:  optional QDialogButtonBox object to receive the help button
    :param button:  optional button to bind directly to the help URL
    :param contextHelp:  flag to determine whether the F1 key is bound for fields 
        with the BoundFieldClass property (default True)
    """
    def __init__(self, parent, url=None, buttonBox=None, button=None, contextHelp=True):
        QtCore.QObject.__init__(self, parent)
        self.parent = parent
        if url is None:
            url = "{0}_Screen.html".format(parent.objectName())
        self.url = url

        if buttonBox is not None:
            self.add_button(buttonBox)
        else:
            self.button = button
            self.button.clicked.connect(self.global_help)

        if contextHelp:
            self.parent.installEventFilter(self)

    def add_button(self, buttonBox):
        self.button = QtGui.QPushButton("Help", self.parent)
        self.button.clicked.connect(self.global_help)
        buttonBox.addButton(self.button, QtGui.QDialogButtonBox.HelpRole)

    def global_help(self):
        from qtalchemy import xplatform
        xplatform.xdg_open(self.url)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.KeyPress and event.key() == QtCore.Qt.Key_F1:
            # iterate from the current focus up the parent chain looking for a 
            # qtalchemy yoke bound control with the BoundClassField property.
            prop = None
            focus = QtGui.QApplication.focusWidget()
            while focus is not None and focus is not self.parent:
                prop = focus.property("BoundClassField")
                if prop is None:
                    focus = focus.parentWidget()
                else:
                    break
            if prop is not None:
                self.context_help(prop)

        return QtCore.QObject.eventFilter(self, obj, event)

    def context_help(self, anchor):
        # reStructuredText anchors are sanitized rather harshly -- we must do likewise here
        anchor = re.sub('[._]', '-', anchor.lower())
        anchored = "{0}#{1}".format(self.url, anchor)
        from qtalchemy import xplatform
        xplatform.xdg_open(anchored)


def message_excepthook(type, value, tb):
    """
    Replace sys.excepthook with this function to display errors more gracefully 
    for an application which is not associated with a console.
    
    Refer to qtalchemy.xplatform.guiexcepthook for a lighter weight version of this.
    
    >>> import sys
    >>> from qtalchemy import *
    >>> sys.excepthook = message_excepthook
    """
    try:
        import traceback

        msgBox = QtGui.QMessageBox()
        msgBox.setWindowTitle(QtGui.QApplication.applicationName())
        msgBox.setIcon(QtGui.QMessageBox.Critical)
        
        tb2 = "An unhandled error was encountered.  Consider copying this entire message to an email to the program author by pressing Ctrl+C on this message.  Include a brief description of the actions leading to this error.\n\n{0}" \
                    .format(''.join(traceback.format_exception(type, value, tb, limit=None)))
        
        msgBox.setText( tb2 )
        
        msgBox.setStandardButtons(QtGui.QMessageBox.Ok)
        msgBox.exec_()
    except:
        #raise
        import sys
        traceback.print_exception(type, value, tb, limit=None, file=sys.stderr)


_test_app = None

def qtapp():
    """
    A QApplication creator for test cases.  QApplication is a single-ton and 
    this provides a safe construction wrapper.
    
    >>> app=qtapp()
    >>> # put test code here
    """
    global _test_app
    _test_app = QtGui.QApplication.instance()
    if _test_app is None:
        _test_app = QtGui.QApplication([])
    return _test_app



# -*- coding: utf-8 -*-
"""
We create a session and model objects and display messages.
"""

from sqlalchemy import create_engine
from qtalchemy import *
from PyQt4 import QtCore

class Vegetable(ModelObject):
    def __init__(self):
        self.msgCounter = 0

    classEvents = ModelObject.Events()

    name = UserAttr(str, "Vegetable Name")
    color = UserAttr(str, "Color")

    @classEvents.add("check", "name")
    def check_name(self, attr, newvalue):
        if newvalue.lower() == "bean":
            raise ValueError("The name 'bean' is not allowed.")

    @classEvents.add("set", "name")
    def set_name(self, attr, oldvalue):
        if self.name.lower() == "tomato":
            self.color = "red"
        if self.name.lower() == "broccoli":
            self.color = "green"

    @classEvents.add("set", "")
    def set_name(self, attr, oldvalue):
        pass
        #print "{0} set to {1}".format(attr, getattr(self, attr))

    @classEvents.add("message")
    def message(self, flags, msg):
        self.msgCounter += 1
        assert msg == "The Message"

def test_class_hooks():
    i = Vegetable()
    try:
        i.name = "bEan"
        assert False, "the checker didn't run"
    except Exception as e:
        pass
    i.name = "Tomato"
    assert i.color == "red", "the set event failed"
    i.modelMessage(Message.Ok|Message.Warning, "The Message")

    assert i.msgCounter == 1, "the message hook wasn't called"

msgCounter = 0
def test_session_hook():
    i = Vegetable()

    Session = PBSessionMaker(bind=create_engine("sqlite://"))

    session = Session()
    session.npadd(i)

    @instanceEvent(session, "message")
    def session_message(self, instance, flags, msg):
        global msgCounter
        msgCounter += 1
        assert msg == "The Message"

    i.modelMessage(Message.Ok|Message.Warning, "The Message")

    assert i.msgCounter == 1, "the message hook wasn't called"
    assert msgCounter == 1, "the message hook wasn't called"
